# Copyright (C) 2020-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import logging as log
import os.path as osp
import random
import re
from collections import Counter, defaultdict
from copy import deepcopy
from enum import Enum, auto
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
import pycocotools.mask as mask_utils
from pandas.api.types import CategoricalDtype

import datumaro.util.mask_tools as mask_tools
from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    Caption,
    Ellipse,
    Label,
    LabelCategories,
    Mask,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RleMask,
    Tabular,
    TabularCategories,
)
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetInfo, DatasetItem, IDataset
from datumaro.components.errors import (
    AnnotationTypeError,
    DatumaroError,
    EmptyCaption,
    EmptyLabel,
    FarFromAttrMean,
    FarFromCaptionMean,
    FarFromLabelMean,
    InvalidValue,
    MissingAnnotation,
    MissingAttribute,
    MissingLabelCategories,
    MultiLabelAnnotations,
    NegativeLength,
    OutlierInCaption,
    RedundanciesInCaption,
    UndefinedAttribute,
    UndefinedLabel,
)
from datumaro.components.media import Image, TableRow
from datumaro.components.transformer import ItemTransform, Transform
from datumaro.util import NOTSET, filter_dict, parse_json_file, parse_str_enum_value, take_by
from datumaro.util.annotation_util import find_group_leader, find_instances
from datumaro.util.tabular_util import emoji_pattern


class CropCoveredSegments(ItemTransform, CliPlugin):
    """
    Sorts polygons and masks ("segments") according to `z_order`,
    crops covered areas of underlying segments. If a segment is split
    into several independent parts by the segments above, produces
    the corresponding number of separate annotations joined into a group.
    """

    def transform_item(self, item):
        annotations = []
        segments = []
        for ann in item.annotations:
            if ann.type in {AnnotationType.polygon, AnnotationType.mask}:
                segments.append(ann)
            else:
                annotations.append(ann)
        if not segments:
            return item

        if not isinstance(item.media, Image):
            raise Exception("Image info is required for this transform")
        h, w = item.media.size
        segments = self.crop_segments(segments, w, h)

        annotations += segments
        return self.wrap_item(item, annotations=annotations)

    @classmethod
    def crop_segments(cls, segment_anns, img_width, img_height):
        segment_anns = sorted(segment_anns, key=lambda x: x.z_order)

        segments = []
        for s in segment_anns:
            if s.type == AnnotationType.polygon:
                segments.append(s.points)
            elif s.type == AnnotationType.mask:
                if isinstance(s, RleMask):
                    rle = s.rle
                else:
                    rle = mask_tools.mask_to_rle(s.image)
                segments.append(rle)

        segments = mask_tools.crop_covered_segments(segments, img_width, img_height)

        new_anns = []
        for ann, new_segment in zip(segment_anns, segments):
            fields = {
                "z_order": ann.z_order,
                "label": ann.label,
                "id": ann.id,
                "group": ann.group,
                "attributes": ann.attributes,
            }
            if ann.type == AnnotationType.polygon:
                if fields["group"] is None:
                    fields["group"] = cls._make_group_id(segment_anns + new_anns, fields["id"])
                for polygon in new_segment:
                    new_anns.append(Polygon(points=polygon, **fields))
            else:
                rle = mask_tools.mask_to_rle(new_segment)
                rle = mask_utils.frPyObjects(rle, *rle["size"])
                new_anns.append(RleMask(rle=rle, **fields))

        return new_anns

    @staticmethod
    def _make_group_id(anns, ann_id):
        if ann_id:
            return ann_id
        max_gid = max(anns, default=0, key=lambda x: x.group)
        return max_gid + 1


class MergeInstanceSegments(ItemTransform, CliPlugin):
    """
    Replaces instance masks and, optionally, polygons with a single mask.
    A group of annotations with the same group id is considered an "instance".
    The largest annotation in the group is considered the group "head", so the
    resulting mask takes properties from that annotation.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--include-polygons", action="store_true", help="Include polygons")
        return parser

    def __init__(self, extractor, include_polygons=False):
        super().__init__(extractor)

        self._include_polygons = include_polygons

    def transform_item(self, item):
        annotations = []
        segments = []
        for ann in item.annotations:
            if ann.type in {AnnotationType.polygon, AnnotationType.mask}:
                segments.append(ann)
            else:
                annotations.append(ann)
        if not segments:
            return item

        if not isinstance(item.media, Image):
            raise Exception("Image info is required for this transform")
        h, w = item.media.size
        instances = self.find_instances(segments)
        segments = [self.merge_segments(i, w, h, self._include_polygons) for i in instances]
        segments = sum(segments, [])

        annotations += segments
        return self.wrap_item(item, annotations=annotations)

    @classmethod
    def merge_segments(cls, instance, img_width, img_height, include_polygons=False):
        polygons = [a for a in instance if a.type == AnnotationType.polygon]
        masks = [a for a in instance if a.type == AnnotationType.mask]
        if not polygons and not masks:
            return []
        if not polygons and len(masks) == 1:
            return masks

        leader = find_group_leader(polygons + masks)
        instance = []

        # Build the resulting mask
        mask = None

        if include_polygons and polygons:
            polygons = [p.points for p in polygons]
            mask = mask_tools.rles_to_mask(polygons, img_width, img_height)
        else:
            instance += polygons  # keep unused polygons

        if masks:
            masks = (m.image for m in masks)
            if mask is not None:
                masks = chain(masks, [mask])
            mask = mask_tools.merge_masks(masks)

        if mask is None:
            return instance

        mask = mask_tools.mask_to_rle(mask)
        mask = mask_utils.frPyObjects(mask, *mask["size"])
        instance.append(
            RleMask(
                rle=mask,
                label=leader.label,
                z_order=leader.z_order,
                id=leader.id,
                attributes=leader.attributes,
                group=leader.group,
            )
        )
        return instance

    @staticmethod
    def find_instances(annotations):
        return find_instances(
            a for a in annotations if a.type in {AnnotationType.polygon, AnnotationType.mask}
        )


class PolygonsToMasks(ItemTransform, CliPlugin):
    _allowed_types = {AnnotationType.polygon, AnnotationType.ellipse}

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type in self._allowed_types:
                if not isinstance(item.media, Image):
                    raise Exception("Image info is required for this transform")
                h, w = item.media.size
                annotations.append(self.convert_polygon(ann, h, w))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_polygon(polygon: Union[Polygon, Ellipse], img_h, img_w):
        rle = mask_utils.frPyObjects([polygon.as_polygon()], img_h, img_w)[0]

        return RleMask(
            rle=rle,
            label=polygon.label,
            z_order=polygon.z_order,
            id=polygon.id,
            attributes=polygon.attributes,
            group=polygon.group,
        )


class BoxesToMasks(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type == AnnotationType.bbox:
                if not isinstance(item.media, Image):
                    raise Exception("Image info is required for this transform")
                h, w = item.media.size
                annotations.append(self.convert_bbox(ann, h, w))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_bbox(bbox, img_h, img_w):
        rle = mask_utils.frPyObjects([bbox.as_polygon()], img_h, img_w)[0]

        return RleMask(
            rle=rle,
            label=bbox.label,
            z_order=bbox.z_order,
            id=bbox.id,
            attributes=bbox.attributes,
            group=bbox.group,
        )


class BoxesToPolygons(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = [
            self.convert_bbox(ann) if ann.type == AnnotationType.bbox else ann
            for ann in item.annotations
        ]

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_bbox(bbox: Bbox):
        return Polygon(
            points=bbox.as_polygon(),
            id=bbox.id,
            attributes=bbox.attributes,
            group=bbox.group,
            label=bbox.label,
            z_order=bbox.z_order,
        )


class MasksToPolygons(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type == AnnotationType.mask:
                polygons = self.convert_mask(ann)
                if not polygons:
                    log.debug(
                        "[%s]: item %s: "
                        "Mask conversion to polygons resulted in too "
                        "small polygons, which were discarded" % (self.NAME, item.id)
                    )
                annotations.extend(polygons)
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_mask(mask):
        polygons = mask_tools.mask_to_polygons(mask.image)

        return [
            Polygon(
                points=p,
                label=mask.label,
                z_order=mask.z_order,
                id=mask.id,
                attributes=mask.attributes,
                group=mask.group,
            )
            for p in polygons
        ]


class ShapesToBoxes(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type in {
                AnnotationType.mask,
                AnnotationType.polygon,
                AnnotationType.polyline,
                AnnotationType.points,
                AnnotationType.ellipse,
            }:
                annotations.append(self.convert_shape(ann))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_shape(shape):
        bbox = shape.get_bbox()
        return Bbox(
            *bbox,
            label=shape.label,
            z_order=shape.z_order,
            id=shape.id,
            attributes=shape.attributes,
            group=shape.group,
        )


class Reindex(Transform, CliPlugin):
    """
    Replaces dataset item IDs with sequential indices.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-s", "--start", type=int, default=1, help="Start value for item ids")
        return parser

    def __init__(self, extractor, start: int = 1):
        super().__init__(extractor)
        self._length = "parent"
        self._start = start

    def __iter__(self):
        for i, item in enumerate(self._extractor):
            yield self.wrap_item(item, id=i + self._start)


class ReindexAnnotations(ItemTransform, CliPlugin):
    """
    Replaces dataset items' annotations with sequential indices.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-s", "--start", type=int, default=1, help="Start value for item ids")
        parser.add_argument(
            "-r",
            "--reindex-each-item",
            action="store_true",
            help="If true, reindex for each item every timeReindex for each item. For example, "
            "item_1 will have ann(id=1), ann(id=2), ..., ann(id=5) and item_2 will have ann_1(id=1), ..., "
            "if reindex_each_item=true and start=1, otherwise, item_2 will have ann_1(id=6), ..., "
            "because item_1 has ann_5(id=5).",
        )
        return parser

    def __init__(self, extractor, start: int = 1, reindex_each_item: bool = False):
        super().__init__(extractor)
        self._length = "parent"
        self._start = start
        self._cur_idx = start
        self._reindex_each_item = reindex_each_item

    def transform_item(self, item: DatasetItem) -> DatasetItem:
        annotations = [
            ann.wrap(id=idx) for idx, ann in enumerate(item.annotations, start=self._cur_idx)
        ]

        self._cur_idx = self._start if self._reindex_each_item else self._cur_idx + len(annotations)

        return item.wrap(annotations=annotations)


class Sort(Transform, CliPlugin):
    """
    Sorts dataset items.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-k", "--key", type=str, default=None, help="key functions to sort.")
        return parser

    def __init__(self, extractor, key=None):
        super().__init__(extractor)
        if key:
            if isinstance(key, str):
                key = eval(key)
            if not callable(key):
                raise Exception("key must be a function with one argument.")
        else:
            key = lambda item: item.id
        self._key = key

    def __iter__(self):
        items = sorted(list(iter(self._extractor)), key=lambda item: self._key(item))
        for item in items:
            yield item


class MapSubsets(ItemTransform, CliPlugin):
    """
    Renames subsets in the dataset.
    """

    @staticmethod
    def _mapping_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return parts

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-s",
            "--subset",
            action="append",
            type=cls._mapping_arg,
            dest="mapping",
            help="Subset mapping of the form: 'src:dst' (repeatable)",
        )
        return parser

    def __init__(self, extractor, mapping=None):
        super().__init__(extractor)

        if mapping is None:
            mapping = {}
        elif not isinstance(mapping, dict):
            mapping = dict(tuple(m) for m in mapping)
        self._mapping = mapping

        if extractor.subsets():
            counts = Counter(mapping.get(s, s) or DEFAULT_SUBSET_NAME for s in extractor.subsets())
            if all(c == 1 for c in counts.values()):
                self._length = "parent"
            self._subsets = set(counts)

    def transform_item(self, item):
        return self.wrap_item(item, subset=self._mapping.get(item.subset, item.subset))


class RandomSplit(Transform, CliPlugin):
    """
    Joins all subsets into one and splits the result into few parts.
    It is expected that item ids are unique and subset ratios sum up to 1.|n
    |n
    Example:|n

    .. code-block::

    |s|s|s|s%(prog)s --subset train:.67 --subset test:.33
    """

    # avoid https://bugs.python.org/issue16399
    _default_split = [("train", 0.67), ("test", 0.33)]

    @staticmethod
    def _split_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return (parts[0], float(parts[1]))

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-s",
            "--subset",
            action="append",
            type=cls._split_arg,
            dest="splits",
            help="Subsets in the form: '<subset>:<ratio>' "
            "(repeatable, default: %s)" % dict(cls._default_split),
        )
        parser.add_argument("--seed", type=int, help="Random seed")
        return parser

    def __init__(self, extractor, splits, seed=None):
        super().__init__(extractor)

        if splits is None:
            splits = self._default_split

        assert 0 < len(splits), "Expected at least one split"
        assert all(0.0 <= r and r <= 1.0 for _, r in splits), (
            "Ratios are expected to be in the range [0; 1], but got %s" % splits
        )

        total_ratio = sum(s[1] for s in splits)
        if not abs(total_ratio - 1.0) <= 1e-7:
            raise Exception(
                "Sum of ratios is expected to be 1, got %s, which is %s" % (splits, total_ratio)
            )

        dataset_size = len(extractor)
        indices = list(range(dataset_size))
        random.seed(seed)
        random.shuffle(indices)
        parts = []
        s = 0
        lower_boundary = 0
        for split_idx, (subset, ratio) in enumerate(splits):
            s += ratio
            upper_boundary = int(s * dataset_size)
            if split_idx == len(splits) - 1:
                upper_boundary = dataset_size
            subset_indices = set(indices[lower_boundary:upper_boundary])
            parts.append((subset_indices, subset))
            lower_boundary = upper_boundary
        self._parts = parts

        self._subsets = set(s[0] for s in splits)
        self._length = "parent"

    def _find_split(self, index):
        for subset_indices, subset in self._parts:
            if index in subset_indices:
                return subset
        return subset  # all the possible remainder goes to the last split

    def __iter__(self):
        for i, item in enumerate(self._extractor):
            yield self.wrap_item(item, subset=self._find_split(i))


class IdFromImageName(ItemTransform, CliPlugin):
    """
    Renames items in the dataset using image file name (without extension).
    """

    def transform_item(self, item):
        if isinstance(item.media, Image) and hasattr(item.media, "path"):
            name = osp.splitext(osp.basename(item.media.path))[0]
            return self.wrap_item(item, id=name)
        else:
            log.debug("Can't change item id for item '%s': " "item has no path info" % item.id)
            return item


class Rename(ItemTransform, CliPlugin):
    r"""
    Renames items in the dataset. Supports regular expressions.
    The first character in the expression is a delimiter for
    the pattern and replacement parts. Replacement part can also
    contain `str.format` replacement fields with the `item`
    (of type `DatasetItem`) object available.|n
    Please use doulbe quotes to represent regex.|n
    |n
    Examples:|n
    |s|s- Replace 'pattern' with 'replacement':|n

      .. code-block::

    |s|s|s|srename -e "|pattern|replacement|"|n
    |n
    |s|s- Remove 'frame_' from item ids:|n

      .. code-block::

    |s|s|s|srename -e "|^frame_||"|n
    |n
    |s|s- Rename by regex:|n

      .. code-block::

    |s|s|s|srename -e "|frame_(\d+)_extra|{item.subset}_id_\1|"
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-e",
            "--regex",
            help="Regex for renaming in the form " "'<sep><search><sep><replacement><sep>'",
        )
        return parser

    def __init__(self, extractor, regex):
        super().__init__(extractor)

        assert regex and isinstance(regex, str)
        parts = regex.split(regex[0], maxsplit=3)
        regex, sub = parts[1:3]
        self._re = re.compile(regex)
        self._sub = sub

    def transform_item(self, item):
        return self.wrap_item(item, id=self._re.sub(self._sub, item.id).format(item=item))


class RemapLabels(ItemTransform, CliPlugin):
    """
    Changes labels in the dataset.|n
    |n
    A label can be:|n
    |s|s- renamed (and joined with existing) -|n
    |s|s|s|swhen '--label <old_name>:<new_name>' is specified|n
    |s|s- deleted - when '--label <name>:' is specified, or default action |n
    |s|s|s|sis 'delete' and the label is not mentioned in the list. |n
    |s|s|s|sWhen a label is deleted, all the associated annotations are removed|n
    |s|s- kept unchanged - when specified '--label <name>:<name>'|n
    |s|s|s|sor default action is 'keep' and the label is not mentioned in the list.|n
    |n
    Annotations with no label are managed by the default action policy.|n
    |n
    Examples:|n
    |n
    |s|s- Remove the 'person' label (and corresponding annotations):|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person: --default keep|n
    |n
    |s|s- Rename 'person' to 'pedestrian' and 'human' to 'pedestrian', join:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person:pedestrian -l human:pedestrian --default keep|n
    |n
    |s|s- Rename 'person' to 'car' and 'cat' to 'dog', keep 'bus', remove others:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person:car -l bus:bus -l cat:dog --default delete
    """

    class DefaultAction(Enum):
        keep = auto()
        delete = auto()

    @staticmethod
    def _split_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return (parts[0], parts[1])

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-l",
            "--label",
            action="append",
            type=cls._split_arg,
            dest="mapping",
            help="Label in the form of: '<src>:<dst>' (repeatable)",
        )
        parser.add_argument(
            "--default",
            choices=[a.name for a in cls.DefaultAction],
            default=cls.DefaultAction.keep.name,
            help="Action for unspecified labels (default: %(default)s)",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        mapping: Union[Dict[str, str], List[Tuple[str, str]]],
        default: Union[None, str, DefaultAction] = None,
    ):
        super().__init__(extractor)

        default = parse_str_enum_value(default, self.DefaultAction, self.DefaultAction.keep)
        self._default_action = default

        assert isinstance(mapping, (dict, list))
        if isinstance(mapping, list):
            mapping = dict(mapping)

        self._categories = {}

        src_categories = self._extractor.categories()

        src_label_cat = src_categories.get(AnnotationType.label)
        if src_label_cat is not None:
            self._make_label_id_map(src_label_cat, mapping, default)

        src_mask_cat = src_categories.get(AnnotationType.mask)
        if src_mask_cat is not None:
            assert src_label_cat is not None
            dst_mask_cat = MaskCategories(attributes=deepcopy(src_mask_cat.attributes))
            for old_id, old_color in src_mask_cat.colormap.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_mask_cat:
                    dst_mask_cat.colormap[new_id] = deepcopy(old_color)

            self._categories[AnnotationType.mask] = dst_mask_cat

        src_point_cat = src_categories.get(AnnotationType.points)
        if src_point_cat is not None:
            assert src_label_cat is not None
            dst_point_cat = PointsCategories(attributes=deepcopy(src_point_cat.attributes))
            for old_id, old_cat in src_point_cat.items.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_point_cat:
                    dst_point_cat.items[new_id] = deepcopy(old_cat)

            self._categories[AnnotationType.points] = dst_point_cat

        assert len(self._categories) == len(src_categories)

    def _make_label_id_map(self, src_label_cat, label_mapping, default_action):
        dst_label_cat = LabelCategories(attributes=deepcopy(src_label_cat.attributes))

        id_mapping = {}
        for src_index, src_label in enumerate(src_label_cat.items):
            dst_label = label_mapping.get(src_label.name, NOTSET)
            if dst_label is NOTSET and default_action == self.DefaultAction.keep:
                dst_label = src_label.name  # keep unspecified as is
            elif not dst_label or dst_label is NOTSET:
                continue

            dst_index = dst_label_cat.find(dst_label)[0]
            if dst_index is None:
                dst_index = dst_label_cat.add(
                    dst_label, src_label.parent, deepcopy(src_label.attributes)
                )
            id_mapping[src_index] = dst_index

        if log.getLogger().isEnabledFor(log.DEBUG):
            log.debug("Label mapping:")
            for src_id, src_label in enumerate(src_label_cat.items):
                if id_mapping.get(src_id) is not None:
                    log.debug(
                        "#%s '%s' -> #%s '%s'",
                        src_id,
                        src_label.name,
                        id_mapping[src_id],
                        dst_label_cat.items[id_mapping[src_id]].name,
                    )
                else:
                    log.debug("#%s '%s' -> <deleted>", src_id, src_label.name)

        self._map_id = lambda src_id: id_mapping.get(src_id, None)

        for label in dst_label_cat:
            if label.parent not in dst_label_cat:
                label.parent = ""
        self._categories[AnnotationType.label] = dst_label_cat

    def categories(self):
        return self._categories

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if getattr(ann, "label", None) is not None:
                conv_label = self._map_id(ann.label)
                if conv_label is not None:
                    annotations.append(ann.wrap(label=conv_label))
            elif self._default_action is self.DefaultAction.keep:
                annotations.append(ann.wrap())
        return item.wrap(annotations=annotations)


class ProjectInfos(Transform, CliPlugin):
    """
    Changes the content of infos.
    A user can add meta-data of dataset such as author, comments, or related papers.
    Infos values are not affect on the dataset structure.
    We thus can add any meta-data freely.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-i",
            "--infos",
            action="append",
            dest="dst_infos",
            help="A dictionary of the dataset meta-information",
        )
        parser.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            dest="overwrite",
            help="Overwrite the infos of src if True",
        )
        return parser

    def __init__(self, extractor: IDataset, dst_infos: DatasetInfo, overwrite: bool = False):
        super().__init__(extractor)

        if overwrite:
            self._infos = dst_infos
        else:
            self._infos = deepcopy(extractor.infos())
            for k, v in dst_infos.items():
                self._infos[k] = v

    def __iter__(self):
        for item in self._extractor:
            if item is not None:
                yield item

    def infos(self):
        return self._infos


class ProjectLabels(ItemTransform):
    """
    Changes the order of labels in the dataset from the existing
    to the desired one, removes unknown labels and adds new labels.
    Updates or removes the corresponding annotations.|n
    |n
    Labels are matched by names (case dependent). Parent labels are only kept
    if they are present in the resulting set of labels. If new labels are
    added, and the dataset has mask colors defined, new labels will obtain
    generated colors.|n
    |n
    Useful for merging similar datasets, whose labels need to be aligned.|n
    |n
    Examples:|n
    |s|s- Align the source dataset labels to [person, cat, dog]:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person -l cat -l dog
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-l",
            "--label",
            action="append",
            dest="dst_labels",
            help="Label name (repeatable, ordered)",
        )
        return parser

    def __init__(self, extractor: IDataset, dst_labels: Union[Iterable[str], LabelCategories]):
        super().__init__(extractor)

        self._categories = {}

        src_categories = self._extractor.categories()

        src_label_cat = src_categories.get(AnnotationType.label)

        if isinstance(dst_labels, LabelCategories):
            dst_label_cat = deepcopy(dst_labels)
        else:
            dst_labels = list(dst_labels)

            if src_label_cat:
                dst_label_cat = LabelCategories(attributes=deepcopy(src_label_cat.attributes))

                for dst_label in dst_labels:
                    assert isinstance(dst_label, str)
                    src_label = src_label_cat.find(dst_label)[1]
                    if src_label is not None:
                        dst_label_cat.add(
                            dst_label, src_label.parent, deepcopy(src_label.attributes)
                        )
                    else:
                        dst_label_cat.add(dst_label)
            else:
                dst_label_cat = LabelCategories.from_iterable(dst_labels)

        for label in dst_label_cat:
            if label.parent not in dst_label_cat:
                label.parent = ""
        self._categories[AnnotationType.label] = dst_label_cat

        self._make_label_id_map(src_label_cat, dst_label_cat)

        src_mask_cat = src_categories.get(AnnotationType.mask)
        if src_mask_cat is not None:
            assert src_label_cat is not None
            dst_mask_cat = MaskCategories(attributes=deepcopy(src_mask_cat.attributes))
            for old_id, old_color in src_mask_cat.colormap.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_mask_cat:
                    dst_mask_cat.colormap[new_id] = deepcopy(old_color)

            # Generate new colors for new labels, keep old untouched
            existing_colors = set(dst_mask_cat.colormap.values())
            color_bank = iter(
                mask_tools.generate_colormap(len(dst_label_cat), include_background=False).values()
            )
            for new_id, new_label in enumerate(dst_label_cat):
                if new_label.name in src_label_cat:
                    continue
                if new_id in dst_mask_cat:
                    continue

                color = next(color_bank)
                while color in existing_colors:
                    color = next(color_bank)

                dst_mask_cat.colormap[new_id] = color

            self._categories[AnnotationType.mask] = dst_mask_cat

        src_point_cat = src_categories.get(AnnotationType.points)
        if src_point_cat is not None:
            assert src_label_cat is not None
            dst_point_cat = PointsCategories(attributes=deepcopy(src_point_cat.attributes))
            for old_id, old_cat in src_point_cat.items.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_point_cat:
                    dst_point_cat.items[new_id] = deepcopy(old_cat)

            self._categories[AnnotationType.points] = dst_point_cat

    def _make_label_id_map(self, src_label_cat, dst_label_cat):
        id_mapping = {
            src_id: dst_label_cat.find(src_label_cat[src_id].name)[0]
            for src_id in range(len(src_label_cat or ()))
        }
        self._map_id = lambda src_id: id_mapping.get(src_id, None)

    def categories(self):
        return self._categories

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if getattr(ann, "label", None) is not None:
                conv_label = self._map_id(ann.label)
                if conv_label is not None:
                    annotations.append(ann.wrap(label=conv_label))
            else:
                annotations.append(ann.wrap())
        return item.wrap(annotations=annotations)


class AnnsToLabels(ItemTransform, CliPlugin):
    """
    Collects all labels from annotations (of all types) and
    transforms them into a set of annotations of type Label
    """

    def transform_item(self, item):
        labels = set(p.label for p in item.annotations if getattr(p, "label") is not None)
        annotations = []
        for label in labels:
            annotations.append(Label(label=label))

        return item.wrap(annotations=annotations)


class BboxValuesDecrement(ItemTransform, CliPlugin):
    """
    Subtracts one from the coordinates of bounding boxes
    """

    def transform_item(self, item):
        annotations = [p for p in item.annotations if p.type != AnnotationType.bbox]
        bboxes = [p for p in item.annotations if p.type == AnnotationType.bbox]
        for bbox in bboxes:
            annotations.append(
                Bbox(
                    bbox.x - 1,
                    bbox.y - 1,
                    bbox.w,
                    bbox.h,
                    label=bbox.label,
                    attributes=bbox.attributes,
                )
            )

        return item.wrap(annotations=annotations)


class ResizeTransform(ItemTransform):
    """
    Resizes images and annotations in the dataset to the specified size.
    Supports upscaling, downscaling and mixed variants.|n
    |n
    Examples:|n
        - Resize all images to 256x256 size|n

        .. code-block::

        |s|s%(prog)s -dw 256 -dh 256
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-dw", "--width", type=int, help="Destination image width")
        parser.add_argument("-dh", "--height", type=int, help="Destination image height")
        return parser

    def __init__(self, extractor: IDataset, width: int, height: int) -> None:
        super().__init__(extractor)

        assert width > 0 and height > 0
        self._width = width
        self._height = height

    @staticmethod
    def _lazy_resize_image(image, new_size):
        def _resize_image():
            h, w = image.size
            yscale = new_size[0] / float(h)
            xscale = new_size[1] / float(w)

            # LANCZOS4 is preferable for upscaling, but it works quite slow
            method = cv2.INTER_AREA if (xscale * yscale) < 1 else cv2.INTER_CUBIC

            resized_image = cv2.resize(image.data / 255.0, new_size[::-1], interpolation=method)
            resized_image *= 255.0
            return resized_image

        return Image.from_numpy(_resize_image, ext=image.ext, size=new_size)

    @staticmethod
    def _lazy_resize_mask(mask, new_size):
        def _resize_image():
            # Can use only NEAREST for masks,
            # because we can't have interpolated values
            rescaled_mask = cv2.resize(
                mask.image.astype(np.float32), new_size[::-1], interpolation=cv2.INTER_NEAREST
            )
            return rescaled_mask.astype(np.uint8)

        return _resize_image

    @staticmethod
    def _lazy_resize_rlemask(mask, new_size):
        def _resize_image():
            # Can use only NEAREST for masks,
            # because we can't have interpolated values
            rescaled_mask = cv2.resize(mask.image, new_size[::-1], interpolation=cv2.INTER_NEAREST)
            return mask_utils.encode(np.asfortranarray(rescaled_mask.astype(np.uint8)))

        return _resize_image

    def transform_item(self, item):
        if not isinstance(item.media, Image):
            raise DatumaroError(
                "Item %s: image info is required for this " "transform" % (item.id,)
            )

        h, w = item.media.size
        xscale = self._width / float(w)
        yscale = self._height / float(h)

        new_size = (self._height, self._width)

        resized_image = None
        if item.media.has_data:
            resized_image = self._lazy_resize_image(item.media, new_size)

        resized_annotations = []
        for ann in item.annotations:
            if isinstance(ann, Bbox):
                resized_annotations.append(
                    ann.wrap(
                        x=ann.x * xscale,
                        y=ann.y * yscale,
                        w=ann.w * xscale,
                        h=ann.h * yscale,
                    )
                )
            elif isinstance(ann, (Polygon, Points, PolyLine)):
                resized_annotations.append(
                    ann.wrap(
                        points=[
                            p
                            for t in ((x * xscale, y * yscale) for x, y in take_by(ann.points, 2))
                            for p in t
                        ]
                    )
                )
            elif isinstance(ann, RleMask):
                rescaled_mask = self._lazy_resize_rlemask(ann, new_size)
                resized_annotations.append(ann.wrap(rle=rescaled_mask))
            elif isinstance(ann, Mask):
                rescaled_mask = self._lazy_resize_mask(ann, new_size)
                resized_annotations.append(ann.wrap(image=rescaled_mask))
            elif isinstance(ann, (Caption, Label)):
                resized_annotations.append(ann)
            else:
                assert False, f"Unexpected annotation type {type(ann)}"

        return self.wrap_item(item, media=resized_image, annotations=resized_annotations)


class RemoveItems(ItemTransform):
    """
    Allows to remove specific dataset items from dataset by their ids.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary samples.|n
    |n
    Examples:|n
        - Remove specific items from the dataset|n

        .. code-block::

        |s|s%(prog)s --id 'image1:train' --id 'image2:test'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) != 2:
            raise argparse.ArgumentTypeError(
                None, message="Invalid id format of '%s'. " "Expected a 'name:subset' pair." % s
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            required=True,
            help="Item id to remove. Id is 'name:subset' pair (repeatable)",
        )
        return parser

    def __init__(self, extractor: IDataset, ids: Iterable[Tuple[str, str]]):
        super().__init__(extractor)
        self._ids = set(tuple(v) for v in (ids or []))

    def transform_item(self, item):
        if (item.id, item.subset) in self._ids:
            return None
        return item


class RemoveAnnotations(ItemTransform):
    """
    Allows to remove annotations on specific dataset items.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary annotations.|n
    |n
    Examples:|n
        - Remove annotations from specific items in the dataset|n

        .. code-block::

        |s|s%(prog)s --id 'image1:train' --id 'image2:test'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) == 3:
            full_id[-1] = int(full_id[-1])
        if len(full_id) != 2 or len(full_id) != 3:
            raise argparse.ArgumentTypeError(
                None,
                message="Invalid id format of '%s'. "
                "Expected 'name:subset:ann_id' or 'name:subset' pair." % s,
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            help="Image id to clean from annotations. "
            "Id is 'name:subset:(optional)ann_id' pair. If ann_id is not specified, "
            "removes all annotations (repeatable) in the item 'name:subset'",
        )
        return parser

    def __init__(self, extractor: IDataset, *, ids: Iterable[Tuple[str, str, Optional[int]]]):
        super().__init__(extractor)

        self._ids = defaultdict(list)
        for v in ids:
            key = tuple(v[:2])
            val = v[2] if len(v) == 3 else None
            if val is not None:
                self._ids[key].append(val)
            else:
                self._ids[key] = []

    def transform_item(self, item: DatasetItem):
        if not self._ids:
            return item

        for item_id, ann_ids in self._ids.items():
            if (item.id, item.subset) == item_id:
                updated_anns = (
                    [ann for ann in item.annotations if ann.id not in ann_ids] if ann_ids else []
                )
                return item.wrap(annotations=updated_anns)

        return item


class RemoveAttributes(ItemTransform):
    """
    Allows to remove item and annotation attributes in a dataset.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary attributes.|n
    |n
    Examples:|n
        - Remove the `is_crowd` attribute from dataset|n

        .. code-block::

        |s|s%(prog)s --attr 'is_crowd'|n
        |n
        - Remove the `occluded` attribute from annotations of|n
        |s|sthe `2010_001705` item in the `train` subset|n

        .. code-block::

        |s|s%(prog)s --id '2010_001705:train' --attr 'occluded'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) != 2:
            raise argparse.ArgumentTypeError(
                None, message="Invalid id format of '%s'. " "Expected a 'name:subset' pair." % s
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            help="Image id to clean from annotations. "
            "Id is 'name:subset' pair. If not specified, "
            "affects all images and annotations (repeatable)",
        )
        parser.add_argument(
            "-a",
            "--attr",
            action="append",
            dest="attributes",
            help="Attribute name to be removed. If not specified, "
            "removes all attributes (repeatable)",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        ids: Optional[Iterable[Tuple[str, str]]] = None,
        attributes: Optional[Iterable[str]] = None,
    ):
        super().__init__(extractor)
        self._ids = set(tuple(v) for v in (ids or []))
        self._attributes = set(attributes or [])

    def _filter_attrs(self, attrs):
        if not self._attributes:
            return None
        else:
            return filter_dict(attrs, exclude_keys=self._attributes)

    def transform_item(self, item: DatasetItem):
        if not self._ids or (item.id, item.subset) in self._ids:
            filtered_annotations = []
            for ann in item.annotations:
                filtered_annotations.append(ann.wrap(attributes=self._filter_attrs(ann.attributes)))

            return item.wrap(
                attributes=self._filter_attrs(item.attributes), annotations=filtered_annotations
            )
        return item


class Correct(Transform, CliPlugin):
    """
    Correct the dataset from a validation report.
    A user can should feed into validation_reports.json from validator to correct the dataset.
    This helps to refine the dataset by rejecting undefined labels, missing annotations, and outliers.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-r",
            "--reports",
            type=str,
            default="validation_reports.json",
            help="A validation report from a 'validate' CLI",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        reports: Union[str, Dict],
    ):
        super().__init__(extractor)

        if isinstance(reports, str):
            reports = parse_json_file(reports)

        self._reports = reports["validation_reports"]

        self._categories = self._extractor.categories()

        self._remove_items = set()
        self._remove_anns = defaultdict(list)
        self._add_attrs = defaultdict(list)

        self._empty_labels = defaultdict(list)
        self._empty_captions = defaultdict(list)
        self._unnecessary_char_captions = list()

        self._outlier_captions = defaultdict(list)
        self._outlier_value = {}

        self._far_from_mean_caption = defaultdict(list)
        self._far_from_mean_value = {}

        self._far_from_mean_caption = defaultdict(list)
        self._far_from_mean_value = {}

        self._analyze_reports(report=self._reports)

        self._table = None
        self._label_value = {}
        self._caption_value = {}

        self.caption_type = {
            cat.name: cat.dtype
            for cat in self._extractor.categories().get(AnnotationType.caption, TabularCategories())
        }

    def categories(self):
        return self._categories

    def _parse_ann_ids(self, desc: str):
        return [int(s) for s in str.split(desc, "'") if s.isdigit()][0]

    def _analyze_reports(self, report):
        for rep in report:
            if rep["anomaly_type"] == MissingLabelCategories.__name__:
                unique_labels = sorted(
                    list({ann.label for item in self._extractor for ann in item.annotations})
                )
                label_categories = LabelCategories().from_iterable(
                    [str(label) for label in unique_labels]
                )
                for item in self._extractor:
                    for ann in item.annotations:
                        attrs = {attr for attr in ann.attributes}
                        label_categories[ann.label].attributes.update(attrs)
                self._categories[AnnotationType.label] = label_categories

            if rep["anomaly_type"] == UndefinedLabel.__name__:
                label_categories = self._categories[AnnotationType.label]
                desc = [s for s in rep["description"].split("'")]
                add_label_name = desc[1]
                label_id, _ = label_categories.find(add_label_name)
                if label_id is None:
                    label_categories.add(name=add_label_name)

            if rep["anomaly_type"] == UndefinedAttribute.__name__:
                label_categories = self._categories[AnnotationType.label]
                desc = [s for s in rep["description"].split("'")]
                attr_name, label_name = desc[1], desc[3]
                label_id = label_categories.find(label_name)[0]
                if label_id is not None:
                    label_categories[label_id].attributes.add(attr_name)

            # [TODO] Correct LabeleDefinedButNotFound: removing a label, reindexing, remapping others
            # if rep["anomaly_type"] == "LabelDefinedButNotFound":
            #     remove_label_name = self._parse_label_cat(rep["description"])
            #     label_cat = self._extractor.categories()[AnnotationType.label]
            #     if remove_label_name in [labels.name for labels in label_cat.items]:
            #         label_cat.remove(remove_label_name)

            if rep["anomaly_type"] in [MissingAnnotation.__name__, MultiLabelAnnotations.__name__]:
                self._remove_items.add((rep["item_id"], rep["subset"]))

            if rep["anomaly_type"] in [
                NegativeLength.__name__,
                InvalidValue.__name__,
                FarFromLabelMean.__name__,
                FarFromAttrMean.__name__,
            ]:
                ann_id = None or self._parse_ann_ids(rep["description"])
                self._remove_anns[(rep["item_id"], rep["subset"])].append(ann_id)

            if rep["anomaly_type"] == MissingAttribute.__name__:
                desc = [s for s in str.split(rep["description"], "'")]
                attr_name, label_name = desc[1], desc[3]
                label_id = self._extractor.categories()[AnnotationType.label].find(label_name)[0]
                self._add_attrs[(rep["item_id"], rep["subset"])].append((label_id, attr_name))

            if rep["anomaly_type"] == RedundanciesInCaption.__name__:
                desc = [s for s in str.split(rep["description"], "'")]
                attr_name, label_name = desc[1], desc[3]
                self._unnecessary_char_captions.append((label_name, attr_name))

            if rep["anomaly_type"] == EmptyLabel.__name__:
                label = rep["description"].split("'")[1]
                self._empty_labels[(rep["item_id"], rep["subset"])].append(label)

            if rep["anomaly_type"] == EmptyCaption.__name__:
                caption = rep["description"].split("'")[1]
                self._empty_captions[(rep["item_id"], rep["subset"])].append(caption)

            if rep["anomaly_type"] == OutlierInCaption.__name__:
                desc = rep["description"].split("'")
                caption = desc[1]
                lower_bound = float(desc[5])
                upper_bound = float(desc[7])
                self._outlier_captions[(rep["item_id"], rep["subset"])].append(caption)
                self._outlier_value[caption] = (lower_bound, upper_bound)

            if rep["anomaly_type"] == FarFromCaptionMean.__name__:
                desc = rep["description"].split("'")
                caption = desc[1]
                lower_bound = float(desc[9])
                upper_bound = float(desc[11])
                self._far_from_mean_caption[(rep["item_id"], rep["subset"])].append(caption)
                self._far_from_mean_value[caption] = (lower_bound, upper_bound)

    def remove_unnecessary_char(self, annotations, item_id):
        if self._table is None:
            items = [item_.media.data() for item_ in self._extractor]
            self._table = pd.DataFrame(items)

        import re

        from nltk.corpus import stopwords

        try:
            stop_words = set(stopwords.words("english"))  # TODO
        except LookupError:
            import nltk

            nltk.download("stopwords")
            stop_words = set(stopwords.words("english"))  # TODO

        def remove_stopwords(text):
            return " ".join([word for word in text.split() if word not in stop_words])

        def remove_urls(text):
            return re.sub(r"http\S+|www\S+|https\S+", "", text, flags=re.MULTILINE)

        def remove_html_tags(text):
            return re.sub(r"<.*?>", "", text)

        def remove_special_characters(text):
            return re.sub(r"[^A-Za-z\s]+", "", text)

        def remove_extra_whitespace(text):
            return re.sub(r"\s+", " ", text).strip()

        def remove_emojis(text):
            return emoji_pattern.sub(r"", text)

        col_redun_dict = {}
        for key, value in self._unnecessary_char_captions:
            if key in col_redun_dict:
                if isinstance(col_redun_dict[key], list):
                    col_redun_dict[key].append(value)
                else:
                    col_redun_dict[key] = [col_redun_dict[key], value]
            else:
                col_redun_dict[key] = [value]

        for col, attr in col_redun_dict.items():
            table = self._table[col].dropna()

            # Remove HTML
            if "html" in attr:
                table = table.apply(remove_html_tags)

            # Remove urls
            if "url" in attr:
                table = table.apply(remove_urls)

            # Remove emojis
            if "emoji" in attr:
                table = table.apply(remove_emojis)

            # Convert to lowercase
            table = table.apply(lambda x: x.lower())

            # Remove Special Characters and Punctuations
            table = table.apply(remove_special_characters)

            # Remove Extra Whitespace
            table = table.apply(remove_extra_whitespace)

            # Remove stopwords
            if "stopword" in attr:
                table = table.apply(remove_stopwords)

        for ann in annotations:
            if ann.type == AnnotationType.caption and ann.caption[: len(col)] == col:
                new_ann = Caption(f"{col}:{table.loc[int(item_id.split('@')[0])]}")
                annotations.remove(ann)
                annotations.append(new_ann)

        return annotations

    def update_caption_value(self):
        if self._table is None:
            items = [item_.media.data() for item_ in self._extractor]
            self._table = pd.DataFrame(items)

        empty_cap_cols = set([v for value in self._empty_captions.values() for v in value])

        for caption in empty_cap_cols:
            table = self._table[caption].dropna()
            caption_type = self.caption_type[caption]

            if caption_type in [int, float]:
                from scipy.stats import skew

                skewness = skew(table)
                if abs(skewness) > 1:  # TODO
                    self._caption_value[caption] = table.median()
                else:
                    self._caption_value[caption] = table.mean()
            elif caption_type is str:
                self._remove_items.update(
                    item_key
                    for item_key, captions in self._empty_captions.items()
                    if caption in captions
                )

    def update_label_value(self):
        if self._table is None:
            items = [item_.media.data() for item_ in self._extractor]
            self._table = pd.DataFrame(items)

        empty_lbl_cols = set([v for value in self._empty_labels.values() for v in value])

        for label in empty_lbl_cols:
            table = self._table[label].dropna()
            # "most_frequency"
            self._label_value[label] = table.mode().iloc[0]

    def fill_missing_value(self, annotations, labels, captions):
        if labels:
            sep_token = self._extractor._sep_token
            id_mapping = self._extractor.categories().get(AnnotationType.label)._indices
            label_value = self._label_value
            annotations.extend(
                Label(id_mapping[f"{label}{sep_token}{label_value[label]}"]) for label in labels
            )
        if captions:
            caption_value = self._caption_value
            annotations.extend(
                Caption(f"{caption}:{caption_value[caption]}") for caption in captions
            )
        return annotations

    def cap_far_from_mean(self, annotations, far_from_mean_captions):
        for ann in annotations:
            if ann.type != AnnotationType.caption:
                continue

            for col in far_from_mean_captions:
                if not ann.caption.startswith(col):
                    continue

                value_str = ann.caption[len(col) + 1 :]
                value = self.caption_type[col](value_str)

                lower_bound, upper_bound = self._far_from_mean_value[col]
                capped_value = max(min(value, upper_bound), lower_bound)

                new_ann = Caption(f"{col}:{capped_value}")
                annotations.remove(ann)
                annotations.append(new_ann)
                break

        return annotations

    def cap_outliers(self, annotations, outliers):
        for ann in annotations:
            if ann.type != AnnotationType.caption:
                continue

            for col in outliers:
                if not ann.caption.startswith(col):
                    continue

                value_str = ann.caption[len(col) + 1 :]
                value = self.caption_type[col](value_str)

                lower_bound, upper_bound = self._outlier_value[col]
                capped_value = max(min(value, upper_bound), lower_bound)

                new_ann = Caption(f"{col}:{capped_value}")
                annotations.remove(ann)
                annotations.append(new_ann)
                break

        return annotations

    def find_outliers(self, annotations, outliers):
        for ann in annotations:
            for col in outliers:
                if ann.type == AnnotationType.caption and ann.caption[: len(col)] == col:
                    value = self._extractor._tabular_cat_types[col](ann.caption[len(col) + 1 :])
                    # Cap outliers
                    lower_bound, upper_bound = self._outlier_value[col]
                    cap_outlier_val = np.where(
                        value > upper_bound,
                        upper_bound,
                        np.where(value < lower_bound, lower_bound, value),
                    ).item()
                    new_ann = Caption(f"{col}:{cap_outlier_val}")
                    annotations.remove(ann)
                    annotations.append(new_ann)
        return annotations

    def __iter__(self):
        if self._empty_captions:
            self.update_caption_value()
        if self._empty_labels:
            self.update_label_value()

        for item in self._extractor:
            item_key = (item.id, item.subset)
            if item_key in self._remove_items:
                continue

            ann_ids = self._remove_anns.get(item_key, None)
            empty_labels = self._empty_labels.get(item_key, None)
            empty_captions = self._empty_captions.get(item_key, None)
            outlier_captions = self._outlier_captions.get(item_key, None)
            far_from_mean_captions = self._far_from_mean_caption.get(item_key, None)

            if ann_ids:
                updated_anns = [ann for ann in item.annotations if ann.id not in ann_ids]
                yield item.wrap(annotations=updated_anns)

            if self._unnecessary_char_captions:
                item.annotations = self.remove_unnecessary_char(item.annotations, item_id=item.id)

            if outlier_captions:
                item.annotations = self.cap_outliers(item.annotations, outlier_captions)

            if far_from_mean_captions:
                item.annotations = self.cap_far_from_mean(item.annotations, far_from_mean_captions)

            if empty_labels or empty_captions:
                item.annotations = self.fill_missing_value(
                    item.annotations, empty_labels, empty_captions
                )

            if not ann_ids:
                updated_attrs = defaultdict(list)
                for label_id, attr_name in self._add_attrs.get((item.id, item.subset), []):
                    updated_attrs[label_id].append(attr_name)

                updated_anns = []
                for ann in item.annotations:
                    new_ann = ann.wrap(attributes=deepcopy(ann.attributes))
                    if ann.type == AnnotationType.label and ann.label in updated_attrs:
                        new_ann.attributes.update(
                            {attr_name: "" for attr_name in updated_attrs[ann.label]}
                        )
                    updated_anns.append(new_ann)
                yield item.wrap(annotations=updated_anns)


class AstypeAnnotations(ItemTransform):
    """
    Enables the conversion of annotation types for the categories and individual items within a dataset.|n
    |n
    Based on a specified mapping, it transforms the annotation types,|n
    changing them to 'Label' if they are categorical, and to 'Caption' if they are of type string, float, or integer.|n
    |n
    Examples:|n
        - Convert type of `title` annotation|n

        .. code-block::

        |s|s%(prog)s --mapping 'title:text'
    """

    @staticmethod
    def _split_arg(s):
        columns = s.split(",")
        results = []
        for column in columns:
            parts = column.split(":")
            if len(parts) != 2:
                raise argparse.ArgumentTypeError()
            results.append((parts[0], parts[1]))
        return results

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--mapping",
            action="append",
            type=cls._split_arg,
            dest="mapping",
            help="Annotations type in the form of: '<src>:<dst>' (repeatable)",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        mapping: Optional[Union[Dict[str, str], List[Tuple[str, str]]]] = None,
    ):
        super().__init__(extractor)

        self._sep_token = ":"

        if extractor.ann_types() and AnnotationType.tabular not in extractor.ann_types():
            raise AnnotationTypeError(
                "Annotation type is not Tabular. This transform only support tabular annotation"
            )

        # Turn off for default setting
        assert mapping is None or isinstance(mapping, (dict, list)), "Mapping must be dict, or list"
        if isinstance(mapping, list):
            mapping = dict(mapping)

        self._categories = {}

        src_categories = self._extractor.categories()
        src_tabular_cat = src_categories.get(AnnotationType.tabular)
        dst_label_cat = LabelCategories()

        if src_tabular_cat is None:
            return

        for src_cat in src_tabular_cat:
            if src_cat.dtype == CategoricalDtype():
                dst_parent = src_cat.name
                dst_labels = sorted(src_cat.labels)
                for dst_label in dst_labels:
                    dst_label = dst_parent + self._sep_token + str(dst_label)
                    dst_label_cat.add(dst_label, parent=dst_parent, attributes={})
                dst_label_cat.add_label_group(src_cat.name, src_cat.labels, group_type=0)
            else:
                self._categories[AnnotationType.caption] = self._categories.get(
                    AnnotationType.caption, []
                ) + [src_cat]
        self._categories[AnnotationType.label] = dst_label_cat

    def categories(self):
        return self._categories

    def transform_item(self, item: DatasetItem):
        if AnnotationType.tabular not in [ann.type for ann in item.annotations]:
            return self.wrap_item(item, annotations=item.annotations)

        label_categories = self._categories.get(AnnotationType.label, LabelCategories())
        labels_set = {item.parent for item in label_categories.items}
        sep_token = self._sep_token
        label_indices = label_categories._indices

        annotations = [
            Label(label=label_indices[f"{name}{sep_token}{value}"])
            if name in labels_set and value is not None
            else Caption(f"{name}{sep_token}{value}")
            for name, value in item.annotations[0].values.items()
            if pd.notna(value)
        ]

        return self.wrap_item(item, annotations=annotations)


class Clean(ItemTransform):
    """
    A class used to refine the media items in a dataset.|n
    |n
    This class provides methods to clean and preprocess media data within a dataset.
    The media data can be of various types such as strings, numeric values, or categorical values.
    The cleaning process for each type of data is handled differently:|n
    |n
    - **String Media**: For string data, the class employs natural language processing (NLP)
    techniques to remove unnecessary characters. This involves cleaning tasks such as removing special
    characters, punctuation, and other irrelevant elements to refine the textual data.|n
    - **Numeric Media**: For numeric data, the class identifies and handles outliers and missing values.
    Outliers are either removed or replaced based on a defined strategy,
    and missing values are filled using appropriate methods such as mean, median, or a predefined value.|n
    """

    def __init__(
        self,
        extractor: IDataset,
    ):
        super().__init__(extractor)

        self._outlier_value = {}
        self._missing_value = {}
        self._sep_token = ":"

    @staticmethod
    def remove_unnecessary_char(text):
        if pd.isna(text):
            return text
        try:
            from nltk.corpus import stopwords

            stop_words = set(stopwords.words("english"))  # TODO
        except LookupError:
            import nltk

            nltk.download("stopwords")
            stop_words = set(stopwords.words("english"))  # TODO

        text = re.sub(r"<.*?>", "", text)  # Remove HTML tags
        text = re.sub(r"http\S+|www\S+|https\S+", "", text, flags=re.MULTILINE)  # remove URLs
        text = emoji_pattern.sub(r"", text)  # Remove emojis
        text = text.lower()  # Convert to lowercase
        text = re.sub(r"[^A-Za-z\s]+", "", text)  # Remove special characters and punctuation
        text = re.sub(r"\s+", " ", text).strip()  # Remove extra whitespaces
        text = " ".join(
            [word for word in text.split() if word not in stop_words]
        )  # Remove stopwords
        return text

    def check_outlier(self, table, numeric_cols):
        for col in numeric_cols:
            col_data = table[col].dropna()

            Q1 = np.quantile(col_data, 0.25)
            Q3 = np.quantile(col_data, 0.75)
            IQR = Q3 - Q1
            lower_bound = Q1 - 1.5 * IQR
            upper_bound = Q3 + 1.5 * IQR

            if table[col].dtype == int:
                lower_bound = self.find_closest_value(col_data[col], lower_bound)
                upper_bound = self.find_closest_value(col_data[col], upper_bound)
            self._outlier_value[col] = (lower_bound, upper_bound)

    def check_missing_value(self, table, float_cols, countable_cols):
        from scipy.stats import skew

        for col in table.columns:
            col_data = table[col].dropna()
            if col in float_cols:
                skewness = skew(col_data)
                if abs(skewness) > 1:  # TODO
                    self._missing_value[col] = col_data.median()
                else:
                    self._missing_value[col] = col_data.mean()
            elif col in countable_cols:
                self._missing_value[col] = col_data.mode().iloc[0]

    @staticmethod
    def find_closest_value(series, target_value):
        abs_diff = np.abs(series - target_value)
        closest_index = abs_diff.idxmin()
        return series.iloc[closest_index]

    def cap_outliers(self, table):
        lower, upper = self._outlier_value[table.name]
        val = table.iloc[0]
        if (val < lower) | (val > upper):
            capped_value = max(min(val, upper), lower)
            return table.replace(val, capped_value)
        return val

    def fill_missing_value(self, series):
        return series.fillna(self._missing_value[series.name])

    def refine_tabular_media(self, item):
        media = item.media
        df = pd.DataFrame(media.data(), index=[media.index])
        str_cols = [col for col in media.data().keys() if media.table.dtype(col) is str]
        float_cols = [col for col in media.data().keys() if media.table.dtype(col) is float]
        int_cols = [col for col in media.data().keys() if media.table.dtype(col) is int]
        countable_cols = [
            col
            for col in media.data().keys()
            if isinstance(item.media.table.dtype(col), CategoricalDtype)
            or item.media.table.dtype(col) is int
        ]

        df[str_cols] = df[str_cols].map(lambda x: self.remove_unnecessary_char(x))

        if not (self._outlier_value):
            self.check_outlier(media.table.data[float_cols + int_cols], float_cols + int_cols)
        df[float_cols + int_cols] = df[float_cols + int_cols].apply(lambda x: self.cap_outliers(x))

        if not (self._missing_value):
            self.check_missing_value(
                media.table.data[float_cols + countable_cols], float_cols, countable_cols
            )
        df[float_cols + countable_cols] = df[float_cols + countable_cols].apply(
            lambda x: self.fill_missing_value(x)
        )

        return TableRow.from_data(
            df.iloc[0].to_dict(), table=item.media.table, index=item.media.index
        )

    def transform_item(self, item):
        if not isinstance(item.media, TableRow):
            raise DatumaroError(
                "Item %s: TableRow info is required for this " "transform" % (item.id,)
            )

        sep_token = self._sep_token
        refined_media = self.refine_tabular_media(item) if item.media.has_data else None
        refined_annotations = []
        for ann in item.annotations:
            if isinstance(ann, Tabular):
                if len(item.annotations) != 1:
                    raise ValueError(
                        "If the item has a tabular annotation, it should have one annotation."
                    )
                annotation_values = {
                    key: refined_media.data[key] for key in item.annotations[0].values.keys()
                }  # only for tabular
                ann = ann.wrap(values=annotation_values)
            elif isinstance(ann, Caption):
                value = [
                    f"{key}{sep_token}{refined_media.data[key]}"
                    for key in refined_media.data.keys()
                    if ann.caption.startswith(key)
                ]
                ann = ann.wrap(caption=value[0])
            refined_annotations.append(ann)

        return self.wrap_item(item, media=refined_media, annotations=refined_annotations)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import Dict, List, Tuple, Type, Union

import numpy as np
import tritonclient.grpc as grpcclient
import tritonclient.http as httpclient

from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred
from datumaro.components.errors import DatumaroError
from datumaro.plugins.inference_server_plugin.base import (
    LauncherForDedicatedInferenceServer,
    ProtocolType,
)

__all__ = ["TritonLauncher"]

TClient = Union[grpcclient.InferenceServerClient, httpclient.InferenceServerClient]
TInferInput = Union[grpcclient.InferInput, httpclient.InferInput]
TInferOutput = Union[grpcclient.InferResult, httpclient.InferResult]


class TritonLauncher(LauncherForDedicatedInferenceServer[TClient]):
    """Inference launcher for Triton Inference Server (https://github.com/triton-inference-server)

    Parameters:
        model_name: Name of the model. It should match with the model name loaded in the server instance.
        model_interpreter_path: Python source code path which implements a model interpreter.
            The model interpreter implement pre-processing of the model input and post-processing of the model output.
        model_version: Version of the model loaded in the server instance
        host: Host address of the server instance
        port: Port number of the server instance
        timeout: Timeout limit during communication between the client and the server instance
        tls_config: Configuration required if the server instance is in the secure mode
        protocol_type: Communication protocol type with the server instance
    """

    def _init_client(self) -> TClient:
        creds = self.tls_config.as_grpc_creds() if self.tls_config is not None else None

        if self.protocol_type == ProtocolType.grpc:
            return grpcclient.InferenceServerClient(url=self.url, creds=creds)
        if self.protocol_type == ProtocolType.http:
            return httpclient.InferenceServerClient(url=self.url)

        raise NotImplementedError(self.protocol_type)

    def _check_server_health(self) -> None:
        status = self._client.is_model_ready(
            model_name=self.model_name,
            model_version=str(self.model_version),
        )
        if not status:
            raise DatumaroError("Model is not ready.")
        log.info(f"Health check succeeded: {status}")

    def _init_metadata(self) -> None:
        self._metadata = self._client.get_model_metadata(
            model_name=self.model_name,
            model_version=str(self.model_version),
        )
        log.info(f"Received metadata: {self._metadata}")

    def _get_infer_input(self, inputs: LauncherInputType) -> TInferInput:
        def _fill_dynamic_axes_dim(
            metadata_shape: Tuple[int, ...], np_data: np.ndarray
        ) -> Tuple[int, ...]:
            """Triton requires to fill the dynamic axes (dim = -1) with the actual dim value of data (>= 0)"""
            if len(metadata_shape) != len(np_data.shape):
                raise ValueError(
                    "Metadata shape and numpy data's shape should be same, "
                    f"but shape ({metadata_shape}) != np_data.shape ({np_data.shape})"
                )

            new_shape = [
                data_dim if metadata_dim == -1 else metadata_dim
                for metadata_dim, data_dim in zip(metadata_shape, np_data.shape)
            ]

            return tuple(new_shape)

        def _get_np_data(input_name: str):
            if isinstance(inputs, np.ndarray):
                return inputs

            if isinstance(inputs, dict):
                np_data = inputs.get(input_name)
                if np_data is None:
                    raise ValueError(f"Input key={input_name} should be given.")
                return np_data

            raise TypeError(inputs)

        def _create(infer_input_cls: Type[TInferInput]) -> TInferInput:
            infer_inputs = [
                infer_input_cls(
                    name=inp.name,
                    shape=_fill_dynamic_axes_dim(inp.shape, _get_np_data(inp.name)),
                    datatype=inp.datatype,
                )
                for inp in self._metadata.inputs
            ]
            for inp in infer_inputs:
                inp.set_data_from_numpy(_get_np_data(inp.name()))
            return infer_inputs

        if self.protocol_type == ProtocolType.grpc:
            return _create(grpcclient.InferInput)
        elif self.protocol_type == ProtocolType.http:
            return _create(httpclient.InferInput)

        raise NotImplementedError(self.protocol_type)

    def infer(self, inputs: LauncherInputType) -> List[ModelPred]:
        infer_outputs: TInferOutput = self._client.infer(
            inputs=self._get_infer_input(inputs),
            model_name=self.model_name,
            model_version=str(self.model_version),
        )

        results: Dict[str, np.ndarray] = {
            output.name: infer_outputs.as_numpy(name=output.name)
            for output in self._metadata.outputs
        }

        outputs_group_by_item = [
            {key: output for key, output in zip(results.keys(), outputs)}
            for outputs in zip(*results.values())
        ]

        return outputs_group_by_item



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import json
import logging as log
import os
from argparse import Namespace
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict

from datumaro.cli.commands.require_project.modification import create, import_

from ...util.errors import CliException
from . import IDatasetDownloader

with open(Path(__file__).parent / "kaggle_formats.json") as f:
    _SUPPORTED_DATASETS = json.load(f)


def make_all_paths_absolute(d: Dict, root: str = "."):
    for k, v in d.items():
        if isinstance(v, dict):
            make_all_paths_absolute(v, root)
        if isinstance(v, str):
            relpath = Path(root) / v
            if relpath.exists():
                d[k] = str(relpath.resolve())


KAGGLE_API_KEY_EXISTS = bool(os.environ.get("KAGGLE_KEY")) or os.path.exists(
    os.path.join(os.path.expanduser("~"), ".kaggle", "kaggle.json")
)


class KaggleDatasetDownloader(IDatasetDownloader):
    @classmethod
    def download(
        cls,
        dataset_id,
        dst_dir=None,
        overwrite=False,
        output_format=None,
        subset=None,
        extra_args=None,
    ):
        try:
            import kaggle
        except ImportError:
            raise CliException(
                "Kaggle API is not installed. To install it, run `pip install kaggle`."
            )

        import_kwargs = _SUPPORTED_DATASETS.get(dataset_id, {})
        if not import_kwargs:
            if not extra_args:
                raise CliException(
                    f"Dataset {dataset_id} has no datumaro-compatible implementation.\n"
                    "Please specify the format and constructor arguments explicitly:\n"
                    "'-- --format=<format> --arg1=<arg1> --arg2=<arg2> ...'"
                )
        else:
            log.info(f"{dataset_id} is supported. Settings:\n{import_kwargs}")
        if output_format:
            log.info(f"Overriding the format with {output_format}...")
            import_kwargs["format"] = output_format
        if "subsets" in import_kwargs:
            if not subset or subset not in import_kwargs["subsets"]:
                raise CliException(
                    f"Please specify the subset. Options : {[k for k in import_kwargs['subsets']]}"
                )
            log.info(f"Getting subset {subset}...")
            import_kwargs = import_kwargs["subsets"][subset]
        if extra_args:
            parsed = import_.build_parser().parse(extra_args)
            format = parsed.format
        else:
            format = import_kwargs.pop("format")

        with TemporaryDirectory() as tmp_dir:
            kaggle.api.dataset_download_cli(dataset_id, path=tmp_dir, force=overwrite, unzip=True)
            create.create_command(Namespace(dst_dir=dst_dir))
            make_all_paths_absolute(import_kwargs, tmp_dir)
            if not extra_args:
                extra_args = {}
                for k, v in import_kwargs.items():
                    if isinstance(v, (dict, list, tuple)):
                        extra_args[k] = json.dumps(v)
                    else:
                        extra_args[k] = str(v)
                extra_args = [f"--{k}={v}" for k, v in extra_args.items()]
            source_name = kaggle.api.split_dataset_string(dataset_id)[1]
            import_.import_command(
                Namespace(
                    _positionals=[],
                    url=tmp_dir,
                    project_dir=dst_dir,
                    format=format,
                    extra_args=extra_args,
                    name=source_name,
                    rpath=None,
                    no_check=False,
                )
            )
            log.info("Dataset downloaded successfully.")

    @classmethod
    def describe(cls, report_format="txt", report_file=None) -> None:
        file = report_file if report_file else None
        if report_format == "txt":
            print(cls.get_command_description(), file=file)

    @classmethod
    def get_command_description(cls) -> str:
        return f"""
Supported datasets: {os.linesep.join(_SUPPORTED_DATASETS)}|n
|n
Examples:|n
- Download the face mask detection dataset:|n
|s|s%(prog)s -i andrewmvd/face-mask-detection"""

    @classmethod
    def describe_command_description(cls):
        return """More detailed
        information can be found in the Kaggle datasets catalog:
        <https://www.kaggle.com/datasets>."""



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import logging as log
import os
from pathlib import Path
from typing import List, Optional, Union

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer, with_subset_dirs
from datumaro.components.media import Image
from datumaro.util.definitions import SUBSET_NAME_BLACKLIST, SUBSET_NAME_WHITELIST
from datumaro.util.image import IMAGE_EXTENSIONS, find_images
from datumaro.util.os_util import walk


class ImagenetPath:
    IMAGE_DIR_NO_LABEL = "no_label"
    SEP_TOKEN = ":"


class ImagenetBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
        min_depth: Optional[int] = None,
        max_depth: Optional[int] = None,
    ):
        if not Path(path).is_dir():
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)
        super().__init__(subset=subset, ctx=ctx)
        self._max_depth = min_depth
        self._min_depth = max_depth
        self._categories = self._load_categories(path)
        self._items = list(self._load_items(path).values())

    def _load_categories(self, path):
        label_cat = LabelCategories()
        path = Path(path)
        for dirname in sorted(d for d in path.rglob("*") if d.is_dir()):
            dirname = dirname.relative_to(path)
            if str(dirname) != ImagenetPath.IMAGE_DIR_NO_LABEL:
                label_cat.add(str(dirname))
        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}

        for image_path in find_images(
            path, recursive=True, max_depth=self._max_depth, min_depth=self._min_depth
        ):
            label = str(Path(image_path).parent.relative_to(path))
            if label == ".":  # image is located in the root directory
                label = ImagenetPath.IMAGE_DIR_NO_LABEL
            image_name = Path(image_path).stem
            item_id = str(label) + ImagenetPath.SEP_TOKEN + image_name
            item = items.get(item_id)
            try:
                if item is None:
                    item = DatasetItem(
                        id=item_id, subset=self._subset, media=Image.from_file(path=image_path)
                    )
                    items[item_id] = item
            except Exception as e:
                self._ctx.error_policy.report_item_error(e, item_id=(item_id, self._subset))
            annotations = item.annotations

            if label != ImagenetPath.IMAGE_DIR_NO_LABEL:
                try:
                    label = self._categories[AnnotationType.label].find(label)[0]
                    annotations.append(Label(label=label))
                    self._ann_types.add(AnnotationType.label)
                except Exception as e:
                    self._ctx.error_policy.report_annotation_error(
                        e, item_id=(item_id, self._subset)
                    )

        return items

    @property
    def is_stream(self) -> bool:
        return True


class ImagenetImporter(Importer):
    """
        Multi-level version of ImagenetImporter.
        For example, it imports the following directory structure.

    .. code-block:: text

        root
        ├── label_0
        │   ├── label_0_1
        │   │   └── img1.jpg
        │   └── label_0_2
        │       └── img2.jpg
        └── label_1
            └── img3.jpg

    """

    _MIN_DEPTH = None
    _MAX_DEPTH = None
    _FORMAT = ImagenetBase.NAME
    DETECT_CONFIDENCE = FormatDetectionConfidence.EXTREME_LOW

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        # Images must not be under a directory whose name is blacklisted.
        for dname, dirnames, filenames in os.walk(context.root_path):
            if dname in SUBSET_NAME_WHITELIST:
                context.fail(
                    f"Following directory names are not permitted: {SUBSET_NAME_WHITELIST}"
                )
            rel_dname = Path(dname).relative_to(context.root_path)
            level = len(rel_dname.parts)
            if cls._MIN_DEPTH is not None and level < cls._MIN_DEPTH and filenames:
                context.fail("Found files out of the directory level bounds.")
            if cls._MAX_DEPTH is not None and level > cls._MAX_DEPTH and filenames:
                context.fail("Found files out of the directory level bounds.")
            dpath = Path(context.root_path) / rel_dname
            if dpath.is_dir():
                if str(rel_dname).lower() in SUBSET_NAME_BLACKLIST:
                    context.fail(
                        f"{dname} is found in {context.root_path}. "
                        "However, Images must not be under a directory whose name is blacklisted "
                        f"(SUBSET_NAME_BLACKLIST={SUBSET_NAME_BLACKLIST})."
                    )

        return super().detect(context)

    @classmethod
    def contains_only_images(cls, path: Union[str, Path]):
        for _, dirnames, filenames in walk(path, cls._MAX_DEPTH, cls._MIN_DEPTH):
            if filenames:
                for filename in filenames:
                    if Path(filename).suffix.lower() not in IMAGE_EXTENSIONS:
                        return False
            elif not dirnames:
                return False
        return True

    @classmethod
    def find_sources(cls, path):
        if not Path(path).is_dir():
            return []

        return [{"url": path, "format": cls._FORMAT}] if cls.contains_only_images(path) else []

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return list(IMAGE_EXTENSIONS)

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--path", required=True)
        parser.add_argument("--subset")

        return parser


@with_subset_dirs
class ImagenetWithSubsetDirsImporter(ImagenetImporter):
    """Multi-level image directory structure importer.
    Example:

    .. code-block::

        root
        ├── train
        │   ├── label_0
        │   │   ├── label_0_1
        │   │   │   └── img1.jpg
        │   │   └── label_0_2
        │   │       └── img2.jpg
        │   └── label_1
        │       └── img3.jpg
        ├── val
        │   ├── label_0
        │   │   ├── label_0_1
        │   │   │   └── img1.jpg
        │   │   └── label_0_2
        │   │       └── img2.jpg
        │   └── label_1
        │       └── img3.jpg
        └── test
            │   ├── label_0
            │   ├── label_0_1
            │   │   └── img1.jpg
            │   └── label_0_2
            │       └── img2.jpg
            └── label_1
                └── img3.jpg
    """


class ImagenetExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"
    USE_SUBSET_DIRS = False

    def _apply_impl(self):
        def _get_name(item: DatasetItem) -> str:
            id_parts = item.id.split(ImagenetPath.SEP_TOKEN)

            if len(id_parts) == 1:
                # e.g. item.id = my_img_1
                return item.id
            else:
                # e.g. item.id = label_1:my_img_1
                return "_".join(id_parts[1:])  # ":" is not allowed in windows

        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        if 1 < len(self._extractor.subsets()) and not self.USE_SUBSET_DIRS:
            log.warning(
                f"There are more than one subset in the dataset ({len(self._extractor.subsets())}). "
                "However, ImageNet format exports all dataset items into the same directory. "
                "Therefore, subset information will be lost. To prevent it, please use ImagenetWithSubsetDirsExporter. "
                'For example, dataset.export("<path/to/output>", format="imagenet_with_subset_dirs").'
            )

        root_dir = Path(self._save_dir)
        extractor = self._extractor
        labels = {}
        for item in self._extractor:
            file_name = _get_name(item)
            labels = set(p.label for p in item.annotations if p.type == AnnotationType.label)

            for label in labels:
                label_name = extractor.categories()[AnnotationType.label][label].name
                self._save_image(
                    item,
                    subdir=root_dir / item.subset / label_name
                    if self.USE_SUBSET_DIRS
                    else root_dir / label_name,
                    name=file_name,
                )

            if not labels:
                self._save_image(
                    item,
                    subdir=root_dir / item.subset / ImagenetPath.IMAGE_DIR_NO_LABEL
                    if self.USE_SUBSET_DIRS
                    else root_dir / ImagenetPath.IMAGE_DIR_NO_LABEL,
                    name=file_name,
                )


class ImagenetWithSubsetDirsExporter(ImagenetExporter):
    USE_SUBSET_DIRS = True



# Copyright (C) 2019-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import logging as log
import os.path as osp
from inspect import isclass
from typing import Any, Dict, Iterator, Optional, Tuple, Type, TypeVar, Union, overload

import pycocotools.mask as mask_utils
from attrs import define

from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    Caption,
    CompiledMask,
    Label,
    LabelCategories,
    Mask,
    Points,
    PointsCategories,
    Polygon,
    RleMask,
)
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetItem, SubsetBase
from datumaro.components.errors import (
    DatasetImportError,
    InvalidAnnotationError,
    InvalidFieldTypeError,
    MissingFieldError,
    UndeclaredLabelError,
)
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image
from datumaro.util import NOTSET, parse_json_file, take_by
from datumaro.util.image import lazy_image, load_image
from datumaro.util.mask_tools import bgr2index
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file

from .format import CocoImporterType, CocoPath, CocoTask
from .page_mapper import COCOPageMapper

T = TypeVar("T")


class DirPathExtracter:
    @staticmethod
    def find_rootpath(path: str) -> str:
        """Find root path from annotation json file path."""
        path = osp.abspath(path)
        if osp.dirname(path).endswith(CocoPath.ANNOTATIONS_DIR):
            return path.rsplit(CocoPath.ANNOTATIONS_DIR, maxsplit=1)[0]
        raise DatasetImportError(
            f"Annotation path ({path}) should be under the directory which is named {CocoPath.ANNOTATIONS_DIR}. "
            "If not, Datumaro fails to find the root path for this dataset. "
            "Please follow this instruction, https://github.com/cocodataset/cocoapi/blob/master/README.txt"
        )

    @staticmethod
    def find_images_dir(rootpath: str, subset: str) -> str:
        """Find images directory from the root path."""

        if rootpath and osp.isdir(osp.join(rootpath, CocoPath.IMAGES_DIR)):
            images_dir = osp.join(rootpath, CocoPath.IMAGES_DIR)
            if osp.isdir(osp.join(images_dir, subset or DEFAULT_SUBSET_NAME)):
                images_dir = osp.join(images_dir, subset or DEFAULT_SUBSET_NAME)
            return images_dir

        raise DatasetImportError(
            f"We found the rootpath ({rootpath}) for this dataset. "
            f"However, there should exist a directory for images as {osp.join(rootpath, CocoPath.IMAGES_DIR)}. "
            "If not, Datumaro fails to find the image directory path. "
            "Please follow this instruction, https://github.com/cocodataset/cocoapi/blob/master/README.txt"
        )


class RoboflowDirPathExtracter(DirPathExtracter):
    @staticmethod
    def find_rootpath(path: str) -> str:
        path, _ = osp.split(path)
        path, _ = osp.split(path)
        return path

    @staticmethod
    def find_images_dir(rootpath: str, subset: str) -> str:
        return osp.join(rootpath, subset)


class MmdetDirPathExtracter(DirPathExtracter):
    @staticmethod
    def find_rootpath(path: str) -> str:
        """Find root path from annotation json file path."""
        path = osp.abspath(path)
        if osp.dirname(path).endswith(CocoPath.ANNOTATIONS_DIR):
            return path.rsplit(CocoPath.ANNOTATIONS_DIR, maxsplit=1)[0]
        raise DatasetImportError(
            f"Annotation path ({path}) should be under the directory which is named {CocoPath.ANNOTATIONS_DIR}. "
            "If not, Datumaro fails to find the root path for this dataset. "
            "Please follow this instruction, https://github.com/cocodataset/cocoapi/blob/master/README.txt"
        )

    @staticmethod
    def find_images_dir(rootpath: str, subset: str) -> str:
        """Find images directory from the root path."""
        return osp.join(rootpath, subset)


class _CocoBase(SubsetBase):
    """
    Parses COCO annotations written in the following format:
    https://cocodataset.org/#format-data
    """

    def __init__(
        self,
        path,
        task,
        *,
        merge_instance_polygons: bool = False,
        keep_original_category_ids: bool = False,
        coco_importer_type: CocoImporterType = CocoImporterType.default,
        subset: Optional[str] = None,
        stream: bool = False,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find JSON file", path)
        self._path = path

        if not subset:
            parts = osp.splitext(osp.basename(path))[0].split(task.name + "_", maxsplit=1)
            subset = parts[1] if len(parts) == 2 else None
        super().__init__(subset=subset, ctx=ctx)

        if coco_importer_type == CocoImporterType.default:
            self._rootpath = DirPathExtracter.find_rootpath(path)
            self._images_dir = DirPathExtracter.find_images_dir(self._rootpath, subset)
        elif coco_importer_type == CocoImporterType.roboflow:
            self._rootpath = RoboflowDirPathExtracter.find_rootpath(path)
            self._images_dir = RoboflowDirPathExtracter.find_images_dir(self._rootpath, subset)
        elif coco_importer_type == CocoImporterType.mmdet:
            self._rootpath = MmdetDirPathExtracter.find_rootpath(path)
            self._images_dir = MmdetDirPathExtracter.find_images_dir(self._rootpath, subset)
        else:
            raise DatasetImportError(f"Not supported type: {coco_importer_type}")

        self._task = task

        self._merge_instance_polygons = merge_instance_polygons

        self._label_map = {}  # coco_id -> dm_id
        if self._task == CocoTask.panoptic:
            self._mask_dir = osp.splitext(path)[0]
        else:
            self._mask_dir = None

        self._stream = stream
        if not stream:
            self._page_mapper = None  # No use in case of stream = False

            json_data = parse_json_file(path)

            self._load_categories(
                json_data,
                keep_original_ids=keep_original_category_ids,
            )

            self._items = self._load_items(json_data)

            del json_data
        else:
            self._page_mapper = COCOPageMapper(path)

            categories_data = self._page_mapper.stream_parse_categories_data()

            self._load_categories(
                {"categories": categories_data},
                keep_original_ids=keep_original_category_ids,
            )

            self._length = None

    def __len__(self) -> int:
        if self.is_stream:
            if self._length is None:
                # Before we actually iterate over the items, we use the length of item page map.
                # It can be different with the actual length,
                # because there is a possiblity that an item cannot be parsed properly.
                return len(self._page_mapper)
            return self._length
        else:
            return len(self._items)

    def __iter__(self) -> Iterator[DatasetItem]:
        if self.is_stream:
            yield from self._stream_items()
        else:
            yield from self._items.values()

    def _load_categories(self, json_data, *, keep_original_ids):
        self._categories = {}

        if has_meta_file(self._rootpath):
            labels = parse_meta_file(self._rootpath).keys()
            self._categories = {AnnotationType.label: LabelCategories.from_iterable(labels)}
            # 0 is reserved for no class
            self._label_map = {i + 1: i for i in range(len(labels))}
        elif self._task in [
            CocoTask.instances,
            CocoTask.labels,
            CocoTask.person_keypoints,
            CocoTask.stuff,
            CocoTask.panoptic,
        ]:
            self._load_label_categories(
                self._parse_field(json_data, "categories", list),
                keep_original_ids=keep_original_ids,
            )

            if self._task == CocoTask.person_keypoints:
                self._load_person_kp_categories(self._parse_field(json_data, "categories", list))

        # informs users if 0 is found as category id sicne 0 is reserved for no class
        found = [
            self._categories[AnnotationType.label][label_id].name
            for cat_id, label_id in self._label_map.items()
            if cat_id == 0
            and self._categories[AnnotationType.label][label_id].name.lower() != "background"
        ]
        if found:
            category_name = found[0]
            log.warning(
                "Category id of '0' is reserved for no class (background) but "
                f"category named '{category_name}' with id of '0' is found in {self._path}. "
                "Please be warned that annotations with category id of '0' would have `None` as label. "
                "(https://openvinotoolkit.github.io/datumaro/latest/docs/explanation/formats/coco.html#import-coco-dataset)"
            )

    def _load_label_categories(self, json_cat, *, keep_original_ids):
        categories = LabelCategories()
        label_map = {}

        cats = sorted(
            (
                {
                    "id": self._parse_field(c, "id", int),
                    "name": self._parse_field(c, "name", str),
                    "supercategory": c.get("supercategory"),
                }
                for c in json_cat
            ),
            key=lambda cat: cat["id"],
        )

        if keep_original_ids:
            for cat in cats:
                label_map[cat["id"]] = cat["id"]

                while len(categories) < cat["id"]:
                    categories.add(f"class-{len(categories)}")

                categories.add(cat["name"], parent=cat.get("supercategory"))
        else:
            for idx, cat in enumerate(cats):
                label_map[cat["id"]] = idx
                categories.add(cat["name"], parent=cat.get("supercategory"))

        self._categories[AnnotationType.label] = categories
        self._label_map = label_map

    def _load_person_kp_categories(self, json_cat):
        categories = PointsCategories()
        for cat in json_cat:
            label_id = self._label_map[self._parse_field(cat, "id", int)]
            categories.add(
                label_id,
                labels=self._parse_field(cat, "keypoints", list),
                joints=self._parse_field(cat, "skeleton", list),
            )

        self._categories[AnnotationType.points] = categories

    def _stream_items(self) -> Iterator[DatasetItem]:
        pbars = self._ctx.progress_reporter
        length = 0

        for img_info, ann_infos in pbars.iter(
            self._page_mapper,
            desc=f"Importing '{self._subset}'",
        ):
            parsed = self._parse_item(img_info)
            if parsed is None:
                continue

            _, item = parsed

            for ann_info in ann_infos:
                self._parse_anns(img_info, ann_info, item)

            yield item
            length += 1
            for ann in item.annotations:
                self._ann_types.add(ann.type)

        self._length = length

    def _parse_anns(self, img_info, ann_info, item):
        try:
            if self._task is not CocoTask.panoptic:
                self._load_annotations(ann_info, img_info, parsed_annotations=item.annotations)
            else:
                self._load_panoptic_ann(ann_info, parsed_annotations=item.annotations)
        except Exception as e:
            self._ctx.error_policy.report_annotation_error(
                e, item_id=(ann_info.get("id", None), self._subset)
            )

    def _load_items(self, json_data):
        pbar = self._ctx.progress_reporter

        def _gen_ann(info_lists):
            while info_lists:
                yield info_lists.pop()

        items = {}
        img_infos = {}
        img_lists = self._parse_field(json_data, "images", list)
        for img_info in _gen_ann(img_lists):
            parsed = self._parse_item(img_info)
            if parsed is None:
                continue

            img_id, item = parsed

            # Store item (DatasetItem) and img_info (Dict) to the integer key dictionary
            items[img_id] = item
            img_infos[img_id] = img_info

        ann_lists = self._parse_field(json_data, "annotations", list)
        for ann_info in pbar.iter(
            _gen_ann(ann_lists),
            desc=f"Importing '{self._subset}'",
            total=len(ann_lists),
        ):
            try:
                img_id = self._parse_field(ann_info, "image_id", int)
                if img_id not in img_infos:
                    log.warn(f"Unknown image id '{img_id}'")
                    continue

                # Retrieve item (DatasetItem) and img_info (Dict) from the integer key dictionary
                item = items[img_id]
                img_info = img_infos[img_id]
                self._parse_anns(img_info, ann_info, item)
                for ann in item.annotations:
                    self._ann_types.add(ann.type)
            except Exception as e:
                self._ctx.error_policy.report_annotation_error(
                    e, item_id=(ann_info.get("id", None), self._subset)
                )

        return items

    def _parse_item(self, img_info: Dict[str, Any]) -> Optional[Tuple[int, DatasetItem]]:
        try:
            img_id = self._parse_field(img_info, "id", int)

            if img_info.get("height") and img_info.get("width"):
                image_size = (
                    self._parse_field(img_info, "height", int),
                    self._parse_field(img_info, "width", int),
                )
            else:
                image_size = None

            file_name = self._parse_field(img_info, "file_name", str)
            return img_id, DatasetItem(
                id=osp.splitext(file_name)[0],
                subset=self._subset,
                media=Image.from_file(path=osp.join(self._images_dir, file_name), size=image_size),
                annotations=[],
                attributes={"id": img_id},
            )
        except Exception as e:
            self._ctx.error_policy.report_item_error(
                e, item_id=(img_info.get("id", None), self._subset)
            )

    def _load_panoptic_ann(self, ann, parsed_annotations=None):
        if parsed_annotations is None:
            parsed_annotations = []

        # For the panoptic task, each annotation struct is a per-image
        # annotation rather than a per-object annotation.
        mask_path = osp.join(self._mask_dir, self._parse_field(ann, "file_name", str))
        mask = lazy_image(mask_path, loader=self._load_pan_mask)
        mask = CompiledMask(instance_mask=mask)
        for segm_info in self._parse_field(ann, "segments_info", list):
            cat_id = self._get_label_id(segm_info)
            segm_id = self._parse_field(segm_info, "id", int)
            attributes = {"is_crowd": bool(self._parse_field(segm_info, "iscrowd", int))}
            parsed_annotations.append(
                Mask(
                    image=mask.lazy_extract(segm_id),
                    label=cat_id,
                    id=segm_id,
                    group=segm_id,
                    attributes=attributes,
                )
            )

        return parsed_annotations

    @staticmethod
    def _load_pan_mask(path):
        mask = load_image(path)
        mask = bgr2index(mask)
        return mask

    @define
    class _lazy_merged_mask:
        segmentation: Any
        h: int
        w: int

        def __call__(self):
            rles = mask_utils.frPyObjects(self.segmentation, self.h, self.w)
            return mask_utils.merge(rles)

    def _get_label_id(self, ann):
        cat_id = self._parse_field(ann, "category_id", int)
        if not cat_id:
            return None

        label_id = self._label_map.get(cat_id)
        if label_id is None:
            raise UndeclaredLabelError(str(cat_id))
        return label_id

    @overload
    def _parse_field(self, ann: Dict[str, Any], key: str, cls: Type[T]) -> T:
        ...

    @overload
    def _parse_field(self, ann: Dict[str, Any], key: str, cls: Tuple[Type, ...]) -> Any:
        ...

    def _parse_field(
        self, ann: Dict[str, Any], key: str, cls: Union[Type[T], Tuple[Type, ...]]
    ) -> Any:
        value = ann.get(key, NOTSET)
        if value is NOTSET:
            raise MissingFieldError(key)
        elif not isinstance(value, cls):
            cls = (cls,) if isclass(cls) else cls
            raise InvalidFieldTypeError(
                key, actual=str(type(value)), expected=tuple(str(t) for t in cls)
            )
        return value

    def _load_annotations(self, ann, image_info=None, parsed_annotations=None):
        if parsed_annotations is None:
            parsed_annotations = []

        ann_id = self._parse_field(ann, "id", int)

        attributes = ann.get("attributes", {})
        if "score" in ann:
            attributes["score"] = self._parse_field(ann, "score", (int, float))

        group = ann_id  # make sure all tasks' annotations are merged

        if (
            self._task is CocoTask.instances
            or self._task is CocoTask.person_keypoints
            or self._task is CocoTask.stuff
        ):
            label_id = self._get_label_id(ann)

            attributes["is_crowd"] = bool(self._parse_field(ann, "iscrowd", int))

            if self._task is CocoTask.person_keypoints:
                keypoints = self._parse_field(ann, "keypoints", list)
                if len(keypoints) % 3 != 0:
                    raise InvalidAnnotationError(
                        f"Keypoints have invalid value count {len(keypoints)}, "
                        "which is not divisible by 3. Expected (x, y, visibility) triplets."
                    )

                points = []
                visibility = []
                for x, y, v in take_by(keypoints, 3):
                    points.append(x)
                    points.append(y)
                    visibility.append(v)

                parsed_annotations.append(
                    Points(
                        points,
                        visibility,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                    )
                )

            try:
                segmentation = self._parse_field(ann, "segmentation", (list, dict))
            except MissingFieldError as e:
                log.warn(str(e))
                segmentation = None

            if segmentation and segmentation != [[]]:
                rle = None

                if isinstance(segmentation, list):
                    if not self._merge_instance_polygons:
                        # polygon - a single object can consist of multiple parts
                        for polygon_points in segmentation:
                            if len(polygon_points) % 2 != 0:
                                raise InvalidAnnotationError(
                                    f"Polygon has invalid value count {len(polygon_points)}, "
                                    "which is not divisible by 2."
                                )
                            elif len(polygon_points) < 6:
                                raise InvalidAnnotationError(
                                    f"Polygon has invalid value count {len(polygon_points)}. "
                                    "Expected at least 3 (x, y) pairs."
                                )

                            parsed_annotations.append(
                                Polygon(
                                    points=polygon_points,
                                    label=label_id,
                                    id=ann_id,
                                    attributes=attributes,
                                    group=group,
                                )
                            )
                    else:
                        # merge all parts into a single mask RLE
                        img_h = self._parse_field(image_info, "height", int)
                        img_w = self._parse_field(image_info, "width", int)
                        rle = self._lazy_merged_mask(segmentation, img_h, img_w)
                elif isinstance(segmentation["counts"], list):
                    # uncompressed RLE
                    img_h = self._parse_field(image_info, "height", int)
                    img_w = self._parse_field(image_info, "width", int)

                    mask_size = self._parse_field(segmentation, "size", list)
                    if len(mask_size) != 2:
                        raise InvalidAnnotationError(
                            f"Mask size has wrong value count {len(mask_size)}. Expected 2 values."
                        )
                    mask_h, mask_w = mask_size

                    if not ((img_h == mask_h) and (img_w == mask_w)):
                        raise InvalidAnnotationError(
                            "Mask #%s does not match image size: %s vs. %s"
                            % (ann_id, (mask_h, mask_w), (img_h, img_w))
                        )
                    rle = self._lazy_merged_mask([segmentation], mask_h, mask_w)
                else:
                    # compressed RLE
                    rle = segmentation

                if rle:
                    parsed_annotations.append(
                        RleMask(
                            rle=rle, label=label_id, id=ann_id, attributes=attributes, group=group
                        )
                    )

            bbox = self._parse_field(ann, "bbox", list)
            if bbox and len(bbox) > 0:
                if len(bbox) != 4:
                    raise InvalidAnnotationError(
                        f"Bbox has wrong value count {len(bbox)}. Expected 4 values."
                    )
                x, y, w, h = bbox
                parsed_annotations.append(
                    Bbox(
                        x,
                        y,
                        w,
                        h,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                    )
                )

        elif self._task is CocoTask.labels:
            label_id = self._get_label_id(ann)
            parsed_annotations.append(
                Label(label=label_id, id=ann_id, attributes=attributes, group=group)
            )
        elif self._task is CocoTask.captions:
            caption = self._parse_field(ann, "caption", str)
            parsed_annotations.append(
                Caption(caption, id=ann_id, attributes=attributes, group=group)
            )
        else:
            raise NotImplementedError()

        return parsed_annotations

    @property
    def is_stream(self) -> bool:
        return self._stream

    def get_dataset_item(self, item_key: int) -> Optional[DatasetItem]:
        if self.is_stream:
            img_info = self._page_mapper.get_item_dict(item_key)
            ann_infos = self._page_mapper.get_anns_dict(item_key)

            parsed = self._parse_item(img_info)
            if parsed is None:
                return None

            _, item = parsed

            for ann_info in ann_infos:
                self._parse_anns(img_info, ann_info, item)

            return item
        else:
            return self._items[item_key]

    def iter_item_ids(self) -> Iterator[int]:
        if self.is_stream:
            return self._page_mapper.iter_item_ids()
        else:
            return self._items.keys()


class CocoImageInfoBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.image_info
        super().__init__(path, **kwargs)


class CocoCaptionsBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.captions
        super().__init__(path, **kwargs)


class CocoInstancesBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.instances
        super().__init__(path, **kwargs)


class CocoPersonKeypointsBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.person_keypoints
        super().__init__(path, **kwargs)


class CocoLabelsBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.labels
        super().__init__(path, **kwargs)


class CocoPanopticBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.panoptic
        super().__init__(path, **kwargs)


class CocoStuffBase(_CocoBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = CocoTask.stuff
        super().__init__(path, **kwargs)



# Copyright (C) 2023-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import Callable, Optional

import numpy as np

import datumaro.util.mask_tools as mask_tools
from datumaro.components.annotation import AnnotationType
from datumaro.components.dataset import Dataset

TASK_ANN_TYPE = {
    "classification": AnnotationType.label,
    "multilabel_classification": AnnotationType.label,
    "detection": AnnotationType.bbox,
    "instance_segmentation": AnnotationType.polygon,
    "semantic_segmentation": AnnotationType.mask,
    "tabular": [AnnotationType.label, AnnotationType.caption],
}


class FrameworkConverterFactory:
    @staticmethod
    def create_converter(framework):
        if framework == "torch":
            return DmTorchDataset
        elif framework == "tf":
            return DmTfDataset
        else:
            raise ValueError("Unsupported framework")


class FrameworkConverter:
    def __init__(self, dataset, subset, task):
        self._dataset = dataset
        self._subset = subset
        self._task = task

    def to_framework(self, framework, **kwargs):
        converter_cls = FrameworkConverterFactory.create_converter(framework)
        return converter_cls(
            dataset=self._dataset,
            subset=self._subset,
            task=self._task,
            **kwargs,
        )


class _MultiFrameworkDataset:
    def __init__(
        self,
        dataset: Dataset,
        subset: str,
        task: str,
    ):
        self.dataset = dataset.get_subset(subset)
        self.task = task
        self.subset = subset
        self._ids = [item.id for item in self.dataset]

    def __len__(self) -> int:
        return len(self.dataset)

    def _gen_item(self, idx: int):
        item = self.dataset.get(id=self._ids[idx], subset=self.subset)
        image = item.media.data

        if self.task == "classification":
            label = [ann.label for ann in item.annotations if ann.type == TASK_ANN_TYPE[self.task]]
            if len(label) > 1:
                log.warning(
                    "Item %s has multiple labels and we choose the first one by default. "
                    "Please choose task=multilabel_classification for allowing this",
                    item.id,
                )
            label = label[0]
        elif self.task == "multilabel_classification":
            label = [ann.label for ann in item.annotations if ann.type == TASK_ANN_TYPE[self.task]]
        elif self.task in ["detection", "instance_segmentation"]:
            label = [
                ann.as_dict() for ann in item.annotations if ann.type == TASK_ANN_TYPE[self.task]
            ]
        elif self.task == "semantic_segmentation":
            masks = [
                (ann.image, ann.label)
                for ann in item.annotations
                if ann.type == TASK_ANN_TYPE[self.task]
            ]
            label = mask_tools.merge_masks((mask, label_id) for mask, label_id in masks)
        elif self.task == "tabular":
            label = [
                ann.as_dict() for ann in item.annotations if ann.type in TASK_ANN_TYPE[self.task]
            ]
        return image, label


try:
    import torch

    class DmTorchDataset(_MultiFrameworkDataset, torch.utils.data.Dataset):
        def __init__(
            self,
            dataset: Dataset,
            subset: str,
            task: str,
            transform: Optional[Callable] = None,
            target_transform: Optional[Callable] = None,
            target: Optional[str] = None,
            tokenizer: Optional[tuple[Callable, Callable]] = None,
            vocab: Optional[tuple[Callable, Callable]] = None,
        ):
            super().__init__(dataset=dataset, subset=subset, task=task)

            self.transform = transform
            self.target_transform = target_transform

            if self.task == "tabular":
                if not isinstance(target, dict):
                    raise ValueError(
                        "Target should be a dictionary with 'input' and 'output' keys."
                    )
                self.input_target = target.get("input")
                self.output_target = target.get("output")
                if not self.input_target:
                    raise ValueError(
                        "Please provide target column for tabular task which is used for input"
                    )

                if not (tokenizer and vocab):
                    raise ValueError("Both tokenizer and vocab must be provided for tabular task")
                self.tokenizer = tokenizer
                self.vocab = vocab

        def __getitem__(self, idx):
            image, label = self._gen_item(idx)

            if self.task == "tabular":
                text = image()[self.input_target]

                if self.output_target:
                    src_tokenizer, tgt_tokenizer = self.tokenizer
                    src_vocab, tgt_vocab = self.vocab
                    src_tokens = src_tokenizer(text)
                    src_token_ids = src_vocab(src_tokens)

                    label_text = label[0]["caption"].split(f"{self.output_target}:")[-1]
                    tgt_tokens = tgt_tokenizer(label_text)
                    tgt_token_ids = tgt_vocab(tgt_tokens)

                    return torch.tensor(src_token_ids, dtype=torch.long), torch.tensor(
                        tgt_token_ids, dtype=torch.long
                    )
                else:
                    tokens = self.tokenizer(text)
                    token_ids = self.vocab(tokens)
                    return torch.tensor(token_ids, dtype=torch.long), torch.tensor(
                        label[0]["label"], dtype=torch.long
                    )

            if len(image.shape) == 2:
                image = np.expand_dims(image, axis=-1)

            if self.transform:
                image = self.transform(image)

            if self.target_transform:
                label = self.target_transform(label)

            return image, label

except ImportError:

    class DmTorchDataset:
        def __init__(self):
            raise ImportError("PyTorch package not found. Cannot convert to PyTorch dataset.")


try:
    import tensorflow as tf

    class DmTfDataset(_MultiFrameworkDataset):
        def __init__(
            self,
            dataset: Dataset,
            subset: str,
            task: str,
            output_signature: Optional[tuple] = None,
        ):
            super().__init__(dataset=dataset, subset=subset, task=task)
            self.output_signature = output_signature

        def _get_rawitem(self, item_id):
            item_id = item_id.numpy() if isinstance(item_id, tf.Tensor) else item_id
            image, label = self._gen_item(item_id)

            if len(self.output_signature.keys()) == 2:
                return image, label

            outputs = []
            for key, spec in self.output_signature.items():
                if key == "image":
                    outputs += [image]
                else:
                    outputs += [tf.convert_to_tensor([l.get(spec.name, None) for l in label])]

            return outputs

        def _process_item(self, item_id):
            output_types = [spec.dtype for spec in self.output_signature.values()]
            outputs = tf.py_function(func=self._get_rawitem, inp=[item_id], Tout=output_types)
            output_dict = {}
            for idx, key in enumerate(self.output_signature.keys()):
                output_dict[key] = outputs[idx]

            return output_dict

        def create(self) -> tf.data.Dataset:
            tf_dataset = tf.data.Dataset.range(len(self.dataset)).map(self._process_item)
            tf_dataset = tf_dataset
            return tf_dataset

        def repeat(self, count=None) -> tf.data.Dataset:
            return self.create().repeat(count)

        def batch(self, batch_size, drop_remainder=False) -> tf.data.Dataset:
            return self.create().batch(batch_size, drop_remainder=drop_remainder)

except ImportError:

    class DmTfDataset:
        def __init__(self):
            raise ImportError("Tensorflow package not found. Cannot convert to Tensorflow dataset.")



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from enum import Enum, auto


class DetectionApiPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"

    DEFAULT_IMAGE_EXT = ".jpg"
    IMAGE_EXT_FORMAT = {".jpg": "jpeg", ".jpeg": "jpeg", ".png": "png"}

    LABELMAP_FILE = "label_map.pbtxt"


class TfrecordImporterType(Enum):
    default = auto()
    roboflow = auto()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Optional, Union

import numpy as np
from attr import attrib, attrs

from datumaro.components.abstracts import IMergerContext
from datumaro.components.abstracts.merger import IMatcherContext
from datumaro.components.annotation import Annotation, Points
from datumaro.util.annotation_util import (
    OKS,
    approximate_line,
    bbox_iou,
    max_bbox,
    mean_bbox,
    segment_iou,
)

__all__ = [
    "match_segments_pair",
    "match_segments_more_than_pair",
    "AnnotationMatcher",
    "LabelMatcher",
    "ShapeMatcher",
    "BboxMatcher",
    "PolygonMatcher",
    "MaskMatcher",
    "PointsMatcher",
    "LineMatcher",
    "CaptionsMatcher",
    "Cuboid3dMatcher",
    "ImageAnnotationMatcher",
    "HashKeyMatcher",
    "FeatureVectorMatcher",
]


def match_segments_pair(
    a_segms,
    b_segms,
    distance=segment_iou,
    dist_thresh=1.0,
    label_matcher=lambda a, b: a.label == b.label,
):
    """Match segments and return pairs of the two matched segments"""

    assert callable(distance), distance
    assert callable(label_matcher), label_matcher

    a_segms.sort(key=lambda ann: 1 - ann.attributes.get("score", 1))
    b_segms.sort(key=lambda ann: 1 - ann.attributes.get("score", 1))

    # a_matches: indices of b_segms matched to a bboxes
    # b_matches: indices of a_segms matched to b bboxes
    a_matches = -np.ones(len(a_segms), dtype=int)
    b_matches = -np.ones(len(b_segms), dtype=int)

    distances = np.array([[distance(a, b) for b in b_segms] for a in a_segms])

    # matches: boxes we succeeded to match completely
    # mispred: boxes we succeeded to match, having label mismatch
    matches = []
    mispred = []

    # It needs len(a_segms) > 0 and len(b_segms) > 0
    if len(b_segms) > 0:
        for a_idx, a_segm in enumerate(a_segms):
            matched_b = -1
            max_dist = -1
            b_indices = np.argsort(
                [not label_matcher(a_segm, b_segm) for b_segm in b_segms], kind="stable"
            )  # prioritize those with same label, keep score order
            for b_idx in b_indices:
                if 0 <= b_matches[b_idx]:  # assign a_segm with max conf
                    continue
                d = distances[a_idx, b_idx]
                if d < dist_thresh or d <= max_dist:
                    continue
                max_dist = d
                matched_b = b_idx

            if matched_b < 0:
                continue
            a_matches[a_idx] = matched_b
            b_matches[matched_b] = a_idx

            b_segm = b_segms[matched_b]

            if label_matcher(a_segm, b_segm):
                matches.append((a_segm, b_segm))
            else:
                mispred.append((a_segm, b_segm))

    # *_umatched: boxes of (*) we failed to match
    a_unmatched = [a_segms[i] for i, m in enumerate(a_matches) if m < 0]
    b_unmatched = [b_segms[i] for i, m in enumerate(b_matches) if m < 0]

    return matches, mispred, a_unmatched, b_unmatched


def match_segments_more_than_pair(
    a_segms,
    b_segms,
    distance=segment_iou,
    dist_thresh=1.0,
    label_matcher=lambda a, b: a.label == b.label,
):
    """Match segments and return sets of the matched segments which can be more than two"""

    assert callable(distance), distance
    assert callable(label_matcher), label_matcher

    # a_matches: indices of b_segms matched to a bboxes
    # b_matches: indices of a_segms matched to b bboxes
    a_matches = -np.ones(len(a_segms), dtype=int)
    b_matches = -np.ones(len(b_segms), dtype=int)

    distances = np.array([[distance(a, b) for b in b_segms] for a in a_segms])

    # matches: boxes we succeeded to match completely
    # mispred: boxes we succeeded to match, having label mismatch
    matches = []
    mispred = []

    # It needs len(a_segms) > 0 and len(b_segms) > 0
    if len(b_segms) > 0:
        for a_idx, a_segm in enumerate(a_segms):
            b_indices = np.argsort(
                [not label_matcher(a_segm, b_segm) for b_segm in b_segms], kind="stable"
            )  # prioritize those with same label, keep score order
            for b_idx in b_indices:
                d = distances[a_idx, b_idx]
                if d < dist_thresh:
                    continue

                a_matches[a_idx] = b_idx
                b_matches[b_idx] = a_idx

                b_segm = b_segms[b_idx]

                if label_matcher(a_segm, b_segm):
                    matches.append((a_segm, b_segm))
                else:
                    mispred.append((a_segm, b_segm))

    # *_umatched: boxes of (*) we failed to match
    a_unmatched = [a_segms[i] for i, m in enumerate(a_matches) if m < 0]
    b_unmatched = [b_segms[i] for i, m in enumerate(b_matches) if m < 0]

    return matches, mispred, a_unmatched, b_unmatched


@attrs(kw_only=True)
class AnnotationMatcher:
    _context: Optional[Union[IMatcherContext, IMergerContext]] = attrib(default=None)

    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class LabelMatcher(AnnotationMatcher):
    def distance(self, a, b):
        a_label = self._context.get_any_label_name(a, a.label)
        b_label = self._context.get_any_label_name(b, b.label)
        return a_label == b_label

    def match_annotations(self, sources):
        return [sum(sources, [])]


@attrs(kw_only=True)
class ShapeMatcher(AnnotationMatcher):
    pairwise_dist = attrib(converter=float, default=0.9)
    cluster_dist = attrib(converter=float, default=-1.0)
    _match_segments = attrib(default=match_segments_pair)

    def match_annotations(self, sources: List[List[Annotation]]) -> List[List[Annotation]]:
        distance = self.distance
        label_matcher = self.label_matcher
        pairwise_dist = self.pairwise_dist
        cluster_dist = self.cluster_dist

        if cluster_dist < 0:
            cluster_dist = pairwise_dist

        id_segm = {id(a): (a, id(s)) for s in sources for a in s}

        def _is_close_enough(cluster, extra_id):
            # check if whole cluster IoU will not be broken
            # when this segment is added
            b = id_segm[extra_id][0]
            for a_id in cluster:
                a = id_segm[a_id][0]
                if distance(a, b) < cluster_dist:
                    return False
            return True

        def _has_same_source(cluster, extra_id):
            b = id_segm[extra_id][1]
            for a_id in cluster:
                a = id_segm[a_id][1]
                if a == b:
                    return True
            return False

        # match segments in sources, pairwise
        adjacent = {i: [] for i in id_segm}  # id(sgm) -> [id(adj_sgm1), ...]
        for a_idx, src_a in enumerate(sources):
            for src_b in sources[a_idx + 1 :]:
                matches, _, _, _ = self._match_segments(
                    src_a,
                    src_b,
                    dist_thresh=pairwise_dist,
                    distance=distance,
                    label_matcher=label_matcher,
                )
                for a, b in matches:
                    adjacent[id(a)].append(id(b))

        # join all segments into matching clusters
        clusters = []
        visited = set()
        for cluster_idx in adjacent:
            if cluster_idx in visited:
                continue

            cluster = set()
            to_visit = {cluster_idx}
            while to_visit:
                c = to_visit.pop()
                cluster.add(c)
                visited.add(c)

                for i in adjacent[c]:
                    if i in visited:
                        continue
                    if 0 < cluster_dist and not _is_close_enough(cluster, i):
                        continue
                    if _has_same_source(cluster, i):
                        continue

                    to_visit.add(i)

            clusters.append([id_segm[i][0] for i in cluster])

        return clusters

    def distance(self, a, b):
        return segment_iou(a, b)

    def label_matcher(self, a, b):
        a_label = self._context.get_any_label_name(a, a.label)
        b_label = self._context.get_any_label_name(b, b.label)
        return a_label == b_label


@attrs
class BboxMatcher(ShapeMatcher):
    pass


@attrs
class PolygonMatcher(ShapeMatcher):
    pass


@attrs
class MaskMatcher(ShapeMatcher):
    pass


@attrs(kw_only=True)
class PointsMatcher(ShapeMatcher):
    sigma: Optional[list] = attrib(default=None)
    instance_map = attrib(converter=dict)

    def distance(self, a, b):
        a_bbox = self.instance_map[id(a)][1]
        b_bbox = self.instance_map[id(b)][1]
        if bbox_iou(a_bbox, b_bbox) <= 0:
            return 0
        bbox = mean_bbox([a_bbox, b_bbox])
        return OKS(a, b, sigma=self.sigma, bbox=bbox)


@attrs
class LineMatcher(ShapeMatcher):
    def distance(self, a, b):
        # Compute inter-line area by using the Trapezoid formulae
        # https://en.wikipedia.org/wiki/Trapezoidal_rule
        # Normalize by common bbox and get the bbox fill ratio
        # Call this ratio the "distance"

        # The box area is an early-exit filter for non-intersected figures
        bbox = max_bbox([a, b])
        box_area = bbox[2] * bbox[3]
        if not box_area:
            return 1

        def _approx(line, segments):
            if len(line) // 2 != segments + 1:
                line = approximate_line(line, segments=segments)
            return np.reshape(line, (-1, 2))

        segments = max(len(a.points) // 2, len(b.points) // 2, 5) - 1

        a = _approx(a.points, segments)
        b = _approx(b.points, segments)
        dists = np.linalg.norm(a - b, axis=1)
        dists = dists[:-1] + dists[1:]
        a_steps = np.linalg.norm(a[1:] - a[:-1], axis=1)
        b_steps = np.linalg.norm(b[1:] - b[:-1], axis=1)

        # For the common bbox we can't use
        # - the AABB (axis-alinged bbox) of a point set
        # - the exterior of a point set
        # - the convex hull of a point set
        # because these soultions won't be correctly normalized.
        # The lines can have multiple self-intersections, which can give
        # the inter-line area more than internal area of the options above,
        # producing the value of the distance outside of the [0; 1] range.
        #
        # Instead, we can compute the upper boundary for the inter-line
        # area based on the maximum point distance and line length.
        max_area = np.max(dists) * max(np.sum(a_steps), np.sum(b_steps))

        area = np.dot(dists, a_steps + b_steps) * 0.5 * 0.5 / max(max_area, 1.0)

        return abs(1 - area)


@attrs
class CaptionsMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class Cuboid3dMatcher(ShapeMatcher):
    def distance(self, a, b):
        raise NotImplementedError()


@attrs
class ImageAnnotationMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class HashKeyMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class FeatureVectorMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class TabularMatcher(AnnotationMatcher):
    def match_annotations(self, sources):
        raise NotImplementedError()


@attrs
class RotatedBboxMatcher(ShapeMatcher):
    sigma: Optional[list] = attrib(default=None)

    def distance(self, a, b):
        a = Points([p for pt in a.as_polygon() for p in pt])
        b = Points([p for pt in b.as_polygon() for p in pt])

        return OKS(a, b, sigma=self.sigma)



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import glob
import os.path as osp
from typing import List, Optional

import h5py
import numpy as np

from datumaro.components.annotation import AnnotationType, DepthAnnotation
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image


class NyuDepthV2Base(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        super().__init__(subset=subset, ctx=ctx)

        self._items = list(self._load_items(path).values())

    def _load_items(self, path):
        items = {}

        anno_files = glob.glob(osp.join(path, "**", "*.h5"), recursive=True)
        for anno_file in anno_files:
            item_id = osp.splitext(osp.basename(anno_file))[0]
            with h5py.File(anno_file, "r") as f:
                image = np.transpose(f["rgb"], (1, 2, 0))
                depth = f["depth"][:].astype("float16")

            items[item_id] = DatasetItem(
                id=item_id,
                media=Image.from_numpy(data=image),
                annotations=[DepthAnnotation(image=Image.from_numpy(data=depth))],
            )
            self._ann_types.add(AnnotationType.depth_annotation)

        return items


class NyuDepthV2Importer(Importer):
    _FORMAT_EXT = ".h5"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file(f"*{cls._FORMAT_EXT}")

    @classmethod
    def find_sources(cls, path):
        return [{"url": path, "format": "nyu_depth_v2"}]

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._FORMAT_EXT]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import Any, Dict, Iterator, List, Optional, Tuple

from datumaro.rust_api import CocoPageMapper as CocoPageMapperImpl

__all__ = ["COCOPageMapper"]


class COCOPageMapper:
    """Construct page maps for items and annotations from the JSON file,
    which are used for the stream importer.

    It also provides __iter__() to produce item and annotation dictionaries
    in stream manner after constructing the page map.
    """

    def __init__(self, path: str) -> None:
        self._path = path
        self._impl = CocoPageMapperImpl(path)

    def __iter__(self) -> Iterator[Tuple[Dict, List[Dict]]]:
        for item_key in self.iter_item_ids():
            yield self._impl.get_item_dict(item_key), self._impl.get_anns_dict(item_key)

    def get_item_dict(self, item_key: int) -> Optional[Dict]:
        try:
            return self._impl.get_item_dict(item_key)
        except Exception as e:
            log.error(e)
            return None

    def get_anns_dict(self, item_key: int) -> Optional[List[Dict]]:
        try:
            return self._impl.get_anns_dict(item_key)
        except Exception as e:
            log.error(e)
            return None

    def __len__(self) -> int:
        return len(self._impl)

    def iter_item_ids(self) -> Iterator[int]:
        for item_id in self._impl.get_img_ids():
            yield item_id

    def __del__(self):
        pass

    def stream_parse_categories_data(self) -> Dict[str, Any]:
        """Parse "categories" section from the given JSON file using the stream json parser"""
        return self._impl.categories()

    def __reduce__(self):
        return (self.__class__, (self._path,))



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os
import os.path as osp
from enum import Enum, auto
from typing import Iterable, List, Optional, Sequence, Tuple, Union

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import DatasetImportError, InvalidAnnotationError, MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class ImagenetTxtPath:
    LABELS_FILE = "synsets.txt"
    IMAGE_DIR = "images"


class _LabelsSource(Enum):
    file = auto()
    generate = auto()


def _parse_annotation_line(line: str) -> Tuple[str, str, Sequence[int]]:
    item = line.split('"')
    if 1 < len(item):
        if len(item) == 3:
            item_id = item[1]
            item = item[2].split()
            image = item_id + item[0]
            label_ids = [int(id) for id in item[1:]]
        else:
            raise InvalidAnnotationError(
                "Line %s: unexpected number " "of quotes in filename" % line
            )
    else:
        item = line.split()
        item_id = osp.splitext(item[0])[0]
        image = item[0]
        label_ids = [int(id) for id in item[1:]]

    return item_id, image, label_ids


class ImagenetTxtBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
        labels: Union[Iterable[str], str] = _LabelsSource.file.name,
        labels_file: str = ImagenetTxtPath.LABELS_FILE,
        image_dir: Optional[str] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find dataset file", path)

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
        super().__init__(subset=subset, ctx=ctx)

        root_dir = osp.dirname(path)
        if not image_dir:
            image_dir = ImagenetTxtPath.IMAGE_DIR
        self.image_dir = osp.join(root_dir, image_dir)

        self._generate_labels = False

        if isinstance(labels, str):
            labels_source = _LabelsSource[labels]

            if labels_source == _LabelsSource.generate:
                labels = ()
                self._generate_labels = True
            elif labels_source == _LabelsSource.file:
                if has_meta_file(root_dir):
                    labels = parse_meta_file(root_dir).keys()
                else:
                    labels = self._parse_labels(osp.join(root_dir, labels_file))
            else:
                assert False, "Unhandled labels source %s" % labels_source
        else:
            assert all(isinstance(e, str) for e in labels)

        self._categories = self._load_categories(labels)

        self._items = list(self._load_items(path).values())

    @staticmethod
    def _parse_labels(path):
        with open(path, encoding="utf-8") as labels_file:
            return [s.strip() for s in labels_file]

    def _load_categories(self, labels):
        return {AnnotationType.label: LabelCategories.from_iterable(labels)}

    def _load_items(self, path):
        items = {}

        with open(path, encoding="utf-8") as f:
            for line in f:
                item_id, image, label_ids = _parse_annotation_line(line)

                anno = []
                label_categories = self._categories[AnnotationType.label]

                for label in label_ids:
                    if label < 0:
                        raise DatasetImportError(f"Image '{item_id}': invalid label id '{label}'")

                    if len(label_categories) <= label:
                        if self._generate_labels:
                            while len(label_categories) <= label:
                                label_categories.add(f"class-{len(label_categories)}")
                        else:
                            raise DatasetImportError(
                                f"Image '{item_id}': unknown label id '{label}'"
                            )

                    anno.append(Label(label))
                    self._ann_types.add(AnnotationType.label)

                items[item_id] = DatasetItem(
                    id=item_id,
                    subset=self._subset,
                    media=Image.from_file(path=osp.join(self.image_dir, image)),
                    annotations=anno,
                )
        return items


class ImagenetTxtImporter(Importer, CliPlugin):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        annot_path = context.require_file(
            f"*{cls._ANNO_EXT}", exclude_fnames=ImagenetTxtPath.LABELS_FILE
        )

        with context.probe_text_file(
            annot_path,
            "must be an ImageNet-like annotation file",
        ) as f:
            for line in f:
                _, _, label_ids = _parse_annotation_line(line)
                if label_ids:
                    break
            else:
                # If there are no labels in the entire file, it's probably
                # not actually an ImageNet file.
                raise Exception

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--labels",
            choices=_LabelsSource.__members__,
            default=_LabelsSource.file.name,
            help="Where to get label descriptions from (use "
            "'file' to load from the file specified by --labels-file; "
            "'generate' to create generic ones)",
        )
        parser.add_argument(
            "--labels-file",
            default=ImagenetTxtPath.LABELS_FILE,
            help="Path to the file with label descriptions (synsets.txt)",
        )
        return parser

    @classmethod
    def find_sources_with_params(cls, path, **extra_params):
        if "labels" not in extra_params or extra_params["labels"] == _LabelsSource.file.name:
            labels_file_name = osp.basename(
                extra_params.get("labels_file") or ImagenetTxtPath.LABELS_FILE
            )

            def file_filter(p):
                return osp.basename(p) != labels_file_name

        else:
            file_filter = None

        return cls._find_sources_recursive(path, ".txt", "imagenet_txt", file_filter=file_filter)

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class ImagenetTxtExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        subset_dir = self._save_dir
        os.makedirs(subset_dir, exist_ok=True)

        extractor = self._extractor
        for subset_name, subset in self._extractor.subsets().items():
            annotation_file = osp.join(subset_dir, "%s.txt" % subset_name)

            labels = {}
            for item in subset:
                item_id = item.id
                if 1 < len(item_id.split()):
                    item_id = '"' + item_id + '"'
                item_id += self._find_image_ext(item)
                labels[item_id] = set(
                    p.label for p in item.annotations if p.type == AnnotationType.label
                )

                if self._save_media and item.media:
                    self._save_image(item, subdir=ImagenetTxtPath.IMAGE_DIR)

            annotation = ""
            for item_id, item_labels in labels.items():
                annotation += "%s %s\n" % (item_id, " ".join(str(l) for l in item_labels))

            with open(annotation_file, "w", encoding="utf-8") as f:
                f.write(annotation)

        if self._save_dataset_meta:
            self._save_meta_file(subset_dir)
        else:
            labels_file = osp.join(subset_dir, ImagenetTxtPath.LABELS_FILE)
            with open(labels_file, "w", encoding="utf-8") as f:
                f.writelines(
                    l.name + "\n"
                    for l in extractor.categories().get(AnnotationType.label, LabelCategories())
                )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
"""Automatic mask generation using Segment Anything Model"""

import os.path as osp
from typing import List, Optional

import numpy as np

import datumaro.plugins.sam_transforms.interpreters.sam_decoder_for_amg as sam_decoder_for_amg
import datumaro.plugins.sam_transforms.interpreters.sam_encoder as sam_encoder_interp
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.transformer import ModelTransform
from datumaro.plugins.inference_server_plugin import OVMSLauncher, TritonLauncher
from datumaro.plugins.inference_server_plugin.base import (
    InferenceServerType,
    ProtocolType,
    TLSConfig,
)
from datumaro.plugins.sam_transforms.interpreters.sam_decoder_for_amg import AMGMasks, AMGPoints

__all__ = ["SAMAutomaticMaskGeneration"]


class SAMAutomaticMaskGeneration(ModelTransform, CliPlugin):
    """Produce instance segmentation masks automatically using Segment Anything Model (SAM).

    This transform can produce instance segmentation mask annotations for each given image.
    It samples single-point input prompts on a uniform 2D grid over the image.
    For each prompt, SAM can predict multiple masks. After obtaining the mask candidates,
    it post-processes them using the given parameters to improve quality and remove duplicates.

    It uses the Segment Anything Model deployed in the OpenVINO™ Model Server
    or NVIDIA Triton™ Inference Server instance. To launch the server instance,
    please see the guide in this link:
    https://github.com/openvinotoolkit/datumaro/tree/develop/docker/segment-anything/README.md

    Parameters:
        extractor: Dataset to transform
        inference_server_type: Inference server type:
            `InferenceServerType.ovms` or `InferenceServerType.triton`
        host: Host address of the server instance
        port: Port number of the server instance
        timeout: Timeout limit during communication between the client and the server instance
        tls_config: Configuration required if the server instance is in the secure mode
        protocol_type: Communication protocol type with the server instance
        num_workers: The number of worker threads to use for parallel inference.
            Set to 0 for single-process mode. Default is 0.
        points_per_side (int): The number of points to be sampled
            along one side of the image. The total number of points is
            points_per_side**2 on a uniform 2d grid.
        points_per_batch (int): Sets the number of points run simultaneously
            by the model. Higher numbers may be faster but use more GPU memory.
        pred_iou_thresh (float): A filtering threshold in [0,1], using the
            model's predicted mask quality.
        stability_score_thresh (float): A filtering threshold in [0,1], using
            the stability of the mask under changes to the cutoff used to binarize
            the model's mask predictions.
        stability_score_offset (float): The amount to shift the cutoff when
            calculated the stability score.
        box_nms_thresh (float): The box IoU cutoff used by non-maximal
            suppression to filter duplicate masks.
        min_mask_region_area (int): If >0, postprocessing will be applied
            to remove the binary mask which has the number of 1s less than min_mask_region_area.
    """

    def __init__(
        self,
        extractor: IDataset,
        inference_server_type: InferenceServerType = InferenceServerType.ovms,
        host: str = "localhost",
        port: int = 9000,
        timeout: float = 10.0,
        tls_config: Optional[TLSConfig] = None,
        protocol_type: ProtocolType = ProtocolType.grpc,
        num_workers: int = 0,
        points_per_side: int = 32,
        points_per_batch: int = 128,
        mask_threshold: float = 0.0,
        pred_iou_thresh: float = 0.88,
        stability_score_thresh: float = 0.95,
        stability_score_offset: float = 1.0,
        box_nms_thresh: float = 0.7,
        min_mask_region_area: int = 0,
    ):
        if inference_server_type == InferenceServerType.ovms:
            launcher_cls = OVMSLauncher
        elif inference_server_type == InferenceServerType.triton:
            launcher_cls = TritonLauncher
        else:
            raise ValueError(inference_server_type)

        self._sam_encoder_launcher = launcher_cls(
            model_name="sam_encoder",
            model_interpreter_path=osp.abspath(sam_encoder_interp.__file__),
            model_version=1,
            host=host,
            port=port,
            timeout=timeout,
            tls_config=tls_config,
            protocol_type=protocol_type,
        )
        self._sam_decoder_launcher = launcher_cls(
            model_name="sam_decoder",
            model_interpreter_path=osp.abspath(sam_decoder_for_amg.__file__),
            model_version=1,
            host=host,
            port=port,
            timeout=timeout,
            tls_config=tls_config,
            protocol_type=protocol_type,
        )

        self.points_per_side = points_per_side
        self.points_per_batch = points_per_batch
        self.mask_threshold = mask_threshold
        self.pred_iou_thresh = pred_iou_thresh
        self.stability_score_offset = stability_score_offset
        self.stability_score_thresh = stability_score_thresh
        self.box_nms_thresh = box_nms_thresh
        self.min_mask_region_area = min_mask_region_area

        super().__init__(
            extractor,
            launcher=self._sam_encoder_launcher,
            batch_size=1,
            append_annotation=False,
            num_workers=num_workers,
        )

    @property
    def points_per_side(self) -> int:
        return self._points_per_side

    @points_per_side.setter
    def points_per_side(self, points_per_side: int) -> None:
        points_y = (np.arange(points_per_side) + 0.5) / points_per_side
        points_x = (np.arange(points_per_side) + 0.5) / points_per_side

        points_x = np.tile(points_x[None, :], (points_per_side, 1))
        points_y = np.tile(points_y[:, None], (1, points_per_side))
        self._points_grid = np.stack([points_x, points_y], axis=-1).reshape(-1, 2)
        self._points_per_side = points_per_side

    def _process_batch(
        self,
        batch: List[DatasetItem],
    ) -> List[DatasetItem]:
        img_embeds = self._sam_encoder_launcher.launch(
            batch=[item for item in batch if self._sam_encoder_launcher.type_check(item)]
        )

        items = []
        for item, img_embed in zip(batch, img_embeds):
            amg_masks: List[AMGMasks] = []

            for i in range(0, len(self._points_grid), self.points_per_batch):
                amg_points = [AMGPoints(points=self._points_grid[i : i + self.points_per_batch])]
                item_to_decode = item.wrap(annotations=amg_points + img_embed)

                # Nested list of mask [[mask_0, ...]]
                nested_masks: List[List[AMGMasks]] = self._sam_decoder_launcher.launch(
                    [item_to_decode],
                    stack=False,
                )
                amg_masks += nested_masks[0]

            mask_anns = AMGMasks.cat(amg_masks).postprocess(
                mask_threshold=self.mask_threshold,
                pred_iou_thresh=self.pred_iou_thresh,
                stability_score_offset=self.stability_score_offset,
                stability_score_thresh=self.stability_score_thresh,
                box_nms_thresh=self.box_nms_thresh,
                min_mask_region_area=self.min_mask_region_area,
            )

            items.append(item.wrap(annotations=mask_anns))

        return items



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
"""Transforms using Segment Anything Model"""
from .automatic_mask_gen import SAMAutomaticMaskGeneration
from .bbox_to_inst_mask import SAMBboxToInstanceMask

__all__ = ["SAMBboxToInstanceMask", "SAMAutomaticMaskGeneration"]



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse

from datumaro.util.scope import scope_add, scoped

from ..commands.require_project.modification import add, import_, remove
from ..util import MultilineFormatter, add_subparser
from ..util.project import load_project


def build_info_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor()

    parser.add_argument("name", nargs="?", help="Source name")
    parser.add_argument("-v", "--verbose", action="store_true", help="Show details")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=info_command)

    return parser


def get_info_sensitive_args():
    return {
        info_command: [
            "name",
            "project_dir",
        ],
    }


@scoped
def info_command(args):
    project = scope_add(load_project(args.project_dir))

    if args.name:
        source = project.working_tree.sources[args.name]
        print(source)
    else:
        for name, conf in project.working_tree.sources.items():
            print(name)
            if args.verbose:
                print(conf)


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        description="""
            Manipulate data sources inside of a project.|n
            |n
            A data source is a source of data for a project.
            The project combines multiple data sources into one dataset.
            The role of a data source is to provide dataset items - images
            and/or annotations.|n
            |n
            By default, the project to be operated on is searched for
            in the current directory. An additional '-p' argument can be
            passed to specify project location.
        """,
        formatter_class=MultilineFormatter,
    )

    subparsers = parser.add_subparsers()
    add_subparser(subparsers, "import", import_.build_parser)
    add_subparser(subparsers, "add", add.build_parser)
    add_subparser(subparsers, "remove", remove.build_parser)
    # TODO: we should distinguish this with datumaro/cli/project_commands/dataset_operations/info.py
    add_subparser(subparsers, "info", build_info_parser)

    return parser


def get_sensitive_args():
    return {
        **add.get_sensitive_args(),
        **import_.get_sensitive_args(),
        **remove.get_sensitive_args(),
        **get_info_sensitive_args(),
    }
